"""
Karta is a collection of Python modules for handling vector and raster
geospatial data.
"""

__version__ = 0.4
__all__ = ["vector", "raster", "crs", "errors"]

from . import vector
from . import raster
from . import crs
from . import errors
from .vector import *
from .raster import *




""" Convenience functions for XY(Z) tables. """

from functools import reduce
import numpy as np

def distance_xy(A):
    """ Calculate the cumulative distance at each coordinate pair in *A*. """
    d_ = np.sqrt(np.sum((A[1:,:] - A[:-1,:])**2, axis=1))
    d = [0]
    for a in d_:
        d.append(d[-1]+a)
    return np.array(d)

contains = lambda s, S: s in S

def read_xy(fnm, delimiter='', header_rows=0):
    """ Load a flowline file and return a size-2 array of coordinates. """
    notblank = lambda s: len(s.strip()) > 0
    with open(fnm) as f:
        lines = filter(notblank, f.readlines()[header_rows:])

    if delimiter == '':
        for delimiter in (',', '\t', ' '):
            if contains(delimiter, lines[-1]):
                break
            delimiter = None

    data = [[float(num) for num in line.strip().split(delimiter)]
                for line in lines]
    return np.array(data)

loadxy = read_xy
load_xy = read_xy

def write_xy(dat, fnm, delimiter=' ', header=None):
    """ Write table data in array *dat* as a delimited ASCII text file. """
    cat = lambda a,b: str(a) + delimiter + str(b)
    with open(fnm, 'w') as f:
        if header is not None:
            f.write(header+'\n')
        for line in dat:
            s = reduce(cat, line) + '\n'
            f.write(s)
    return

def xyz2array_reg(X, Y, Z):
    """ Return an array from X,Y,Z vectors, assuming gridding is
    regular. """
    xmin = min(X)
    ymin = min(Y)

    nx = sum([y==ymin for y in Y])
    ny = sum([x==xmin for x in X])

    XYZ = [(x,y,z) for x,y,z in zip(X,Y,Z)]
    XYZ.sort(key=lambda a: a[1])
    XYZ.sort(key=lambda a: a[0])
    Zs = [a[2] for a in XYZ]

    A = np.zeros([ny, nx])
    for i in range(ny):
        A[i,:] = Zs[i*nx:(i+1)*nx]

    return A


def array2xyz(A, X, Y):
    """ There are a few occasions when an XYZ list is the proper data
    format. This function converts from an array *A* with coordinates
    defined by *X* and *Y* to a list of (x,y,z) tuples.
    """
    xyz = []
    m, n = A.shape
    for j in range(n):
        for i in range (m):
            xyz.append( (X[i], Y[j], A[i,j]) )
    return xyz




""" VTK file driver based on documentation from
http://vtk.org/VTK/img/file-formats.pdf """

import sys
import operator
from xml.etree import ElementTree as ET
from xml.dom import minidom
from functools import reduce

def mp2vtp(mp_list, f, **kwargs):
    """ Write a list of Multipoint instances to a serial VTK PolyData (.vtp)
    file.

    kwargs are:

        *vbyteorder*        :   'LittleEndian' or 'BigEndian' to force a byte
                                order other than the native order

        *write_pointdata*   :   {bool}, default True, to write point attributes
    """

    vtype = 'PolyData'
    vbyteorder = kwargs.get('byte_order', 'undef')
    if vbyteorder == 'undef':
        if sys.byteorder == 'little':
            vbyteorder = 'LittleEndian'
        else:
            vbyteorder = 'BigEndian'
    if vbyteorder not in ('LittleEndian', 'BigEndian'):
        raise Exception("vbyteorder must be 'LittleEndian' or 'BigEndian'")
    write_pointdata = kwargs.get('write_pointdata', True)

    vversion = str(kwargs.get('version', 0.1))

    xdoc = ET.Element('VTKFile', attrib={'type':vtype,
                                         'version':vversion,
                                         'byte_order':vbyteorder})

    xugrid = ET.SubElement(xdoc, vtype)

    if hasattr(mp_list, "_geotype") and mp_list._geotype == "Multipoint":
        mp_list = [mp_list]

    for mp in mp_list:

        xpiece = ET.Element('Piece', attrib={'NumberOfPoints':str(len(mp)),
                                             'NumberOfVerts':str(len(mp))})
        xugrid.append(xpiece)

        # Point coordinates
        xpts = ET.Element('Points')
        xdarray = ET.Element('DataArray',
                        attrib={'NumberOfComponents':'{0}'.format(mp.rank),
                                'type':'Float32',
                                'format':'ascii'})
        xdarray.text = reduce(lambda a,b: str(a)+' '+str(b),    # Concat points
                        map(lambda pt:                          # Form points
                            reduce(lambda a,b: str(a)+' '+str(b), pt),
                            mp))                                # Concat coords

        xpts.append(xdarray)
        xpiece.append(xpts)

        # Point data
        if False in (a is None for a in mp.data) and write_pointdata:
            if hasattr(mp.data, 'keys'):
                datakeys = mp.data.keys()
                datavals = mp.data.values()
            else:
                datakeys = ['point_data']
                datavals = [mp.data]

            xptdata = ET.Element('PointData')
            for key, vals in zip(datakeys, datavals):
                if isinstance(vals[0], int):
                    dtype = 'Int32'
                elif isinstance(vals[0], float):
                    dtype = 'Float32'
                else:
                    dtype = 'String'
                xdarray = ET.Element('DataArray', attrib={'type':dtype,
                                                          'format':'ascii',
                                                          'Name':str(key)})
                xdarray.text = reduce(lambda a,b: str(a)+' '+str(b), vals)
                xptdata.append(xdarray)

            xpiece.append(xptdata)

        # Cell data
        xcell = ET.Element('Verts')
        connectivity = ET.Element('DataArray', attrib={'type':'Int32',
                                                       'Name':'connectivity',
                                                       'format':'ascii'})
        connectivity.text = reduce(operator.add,
                                [str(a)+' ' for a in range(len(mp))])

        offsets = ET.Element('DataArray', attrib={'type':'Int32',
                                                  'Name':'offsets',
                                                  'format':'ascii'})
        #offsets.text = str(len(mp))
        offsets.text = reduce(operator.add,
                                [str(a+1)+' ' for a in range(len(mp))])


        xcell.append(connectivity)
        xcell.append(offsets)
        xpiece.append(xcell)

    pretty_xml = minidom.parseString(ET.tostring(xdoc)).toprettyxml()

    f.write(pretty_xml)
    return




""" Cython versions of low-level functions for computing vector geometry. """

import numpy as np
cimport numpy as np
from cpython cimport bool
from libc.math cimport atan, sqrt

cpdef double cdot2(tuple v1, tuple v2):
    return v1[0] * v2[0] + v1[1] * v2[1]

cpdef double cross2(tuple pt1, tuple pt2):
    return pt1[0]*pt2[1] - pt1[1]*pt2[0]

def isleft(tuple pt0, tuple pt1, tuple pt2):
    return (pt1[0]-pt0[0])*(pt2[1]-pt0[1]) - (pt1[1]-pt0[1])*(pt2[0]-pt0[0]) > 0.0

def polarangle(tuple pt0, tuple pt1):
    """ Return the polar angle from pt0 to pt1, where we assume pt0.y
    <= pt1.y """
    cdef double pi
    cdef double dx
    pi = 3.141592653589793
    dx = pt1[0] - pt0[0]
    if dx > 0:
        return atan((pt1[1] - pt0[1]) / dx)
    elif dx < 0:
        return atan((pt1[1] - pt0[1]) / dx) + pi
    else:
        return 0.5*pi

cdef inline double dbl_max(double a, double b): return a if a >= b else b
cdef inline double dbl_min(double a, double b): return a if a <= b else b

cdef bool isbetween_inc(double a, double b, double c):
    return dbl_min(a, c) <= b <= dbl_max(a, c)

cdef bool isbetween_incl(double a, double b, double c):
    return dbl_min(a, c) <= b < dbl_max(a, c)

cdef bool isbetween_incr(double a, double b, double c):
    return dbl_min(a, c) < b <= dbl_max(a, c)

ctypedef bool (*isbetween_t)(double, double, double)

def intersection(double x0, double x1, double x2, double x3,
                 double y0, double y1, double y2, double y3):
    """ Return the point of intersection between two line segments. Returns NaN
    if the lines do not intersect.
    """
    cdef double m0, m1
    cdef double x, y

    if x1 != x0:
        m0 = float(y1-y0) / float(x1-x0)
    else:
        m0 = 1e37
    if x3 != x2:
        m1 = float(y3-y2) / float(x3-x2)
    else:
        m1 = 1e37
    if m0 == m1:
        return (np.nan, np.nan)
    x = float(m0*x0 - m1*x2 + y2 - y0) / float(m0 - m1)

    cdef bool iswithinx
    cdef bool iswithiny
    iswithinx = False
    iswithiny = False

    if abs(x1 - x0) >= 1e-15:
        if abs(x3 - x2) >= 1e-15:
            if isbetween_inc(x0, x, x1) and isbetween_inc(x2, x, x3):
                iswithinx = True
        else:
            if abs(x - x2) < 1e-15 and isbetween_inc(x0, x, x1):
                iswithinx = True
    else:
        if abs(x - x0) < 1e-15 and isbetween_inc(x2, x, x3):
            iswithinx = True

    if iswithinx:
        if abs(x-x0) >= 1e-15:
            y = m0 * (x-x0) + y0
        else:
            y = m1 * (x-x2) + y2
        if abs(y1 - y0) >= 1e-15:
            if abs(y3 - y2) >= 1e-15:
                if isbetween_inc(y0, y, y1) and isbetween_inc(y2, y, y3):
                    iswithiny = True
            elif abs(y - y2) < 1e-15 and isbetween_inc(y0, y, y1):
                iswithiny = True
        elif abs(y - y0) < 1e-15 and isbetween_inc(y2, y, y3):
            iswithiny = True

    if iswithinx and iswithiny:
        return (x, y)
    else:
        return (np.nan, np.nan)

def intersects_cn(double xp, double yp, double x0, double x1, double y0, double y1):
    """ Test whether a vertical ray emanating up from a point (xp, yp) crosses
    a line segment [(x0, y0), (x1, y1)]. Used to implement a crossing number
    membership test.
    """
    cdef double m, y

    if x0 != x1:
        m = (y1-y0) / (x1-x0)
    else:
        return False

    y = y0 + m * (xp - x0)

    if y < yp:
        return False

    cdef bool iswithinx
    cdef bool iswithiny
    iswithinx = False
    iswithiny = False

    if m > 0.0 and isbetween_incr(y0, y, y1):
        iswithiny = True
    elif isbetween_incl(y0, y, y1):
        iswithiny = True
    elif abs(y0-y1) < 1e-15 and abs(y-y0) < 1e-15:
        iswithiny = True

    if isbetween_incr(x0, xp, x1):
        iswithinx = True

    return iswithinx and iswithiny

cdef tuple cproj2(tuple u, tuple v):
    cdef double uv, vv
    uv = cdot2(u,v)
    vv = cdot2(v,v)
    return (uv / vv * v[0], uv / vv * v[1])

def distance(tuple pt0, tuple pt1):
    """ Calculate the distance between two points (tuples) """
    cdef float dsq
    dsq = 0.0
    for i in range(min(len(pt0), len(pt1))):
        dsq += abs(pt0[i] - pt1[i])**2
    return sqrt(dsq)

def pt_nearest_planar(tuple pt, tuple endpt1, tuple endpt2):
    """ Determines the point on a segment defined by tuples endpt1
    and endpt2 nearest to a point defined by tuple pt.
    Returns the a nested tuple of (point, distance)
    """

    cdef tuple u, v, u_int
    cdef tuple u_on_v

    dist = lambda u,v: sqrt((u[0]-v[0])**2. + (u[1]-v[1])**2.)

    u = (pt[0] - endpt1[0], pt[1] - endpt1[1])
    v = (endpt2[0] - endpt1[0], endpt2[1] - endpt1[1])
    u_on_v = cproj2(u,v)
    u_int = (u_on_v[0] + endpt1[0], u_on_v[1] + endpt1[1])

    # Determine whether u_int is inside the segment
    # Otherwise return the nearest endpoint
    if endpt1[0] == endpt2[0]:  # Segment is vertical
        if u_int[1] > max(endpt1[1], endpt2[1]):
            return (endpt1[1] > endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[1] < min(endpt1[1], endpt2[1]):
            return (endpt1[1] < endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))
    else:
        if u_int[0] > max(endpt1[0], endpt2[0]):
            return (endpt1[0] > endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[0] < min(endpt1[0], endpt2[0]):
            return (endpt1[0] < endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))

def pt_nearest_proj(fwd, inv, pt, endpt0, endpt1, float tol=1.0, int maxiter=50):
    """ Given geodetic functions *fwd* and *inv*, a Point *pt*, and an arc from
    *endpt1* to *endpt2*, return the point on the arc that is nearest *pt*.

    Scheme employs a bisection minimization method. Iteration continues until a
    tolerance *tol* in meters is reached, or *maxiter* iterations are
    exhausted. If the iteration limit is reached, a ConvergenceError is raised.
    """
    cdef double az, az2, L
    (az, az2, L) = inv(endpt0[0], endpt0[1], endpt1[0], endpt1[1])

    def distance(double x):
        cdef double trialx, trialy, _, d
        (trialx, trialy, _) = fwd(endpt0[0], endpt0[1], az, x*L)
        (_, _, d) = inv(trialx, trialy, pt[0], pt[1])
        return d

    def ddx(double x):
        cdef double dx, d1, d2
        dx = 1e-8
        d1 = distance(x)
        d2 = distance(x+dx)
        return (d2-d1) / dx

    # Detect whether the nearest point is at an endpoint
    cdef double dx0, dx1
    dx0, dx1 = ddx(0), ddx(1)
    if dx0 > 0:
        return endpt0, distance(0)
    elif dx1 < 0:
        return endpt1, distance(1)

    # Bisection iteration
    cdef float x0, x1, xm, xn, yn
    cdef int i = 0
    cdef float dx = tol + 1.0
    x0, x1 = 0.0, 1.0
    while dx > tol:
        if i == maxiter:
            raise ConvergenceError("Maximum iterations exhausted in bisection "
                                   "method.")
        xm = 0.5 * (x0 + x1)
        if ddx(xm) > 0:
            dx = abs(x1-xm) * L
            x1 = xm
        else:
            dx = abs(x0-xm) * L
            x0 = xm
        i += 1
    (xn, yn, _) = fwd(endpt0[0], endpt0[1], az, xm*L)
    return (xn, yn), distance(xm)

class ConvergenceError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


# QuadTree functions

def iswithin(tuple bbox, tuple pt):
    """ Return whether a point is within a bounding box (planar approximation). """
    cdef double xmn = bbox[0]
    cdef double xmx = bbox[1]
    cdef double ymn = bbox[2]
    cdef double ymx = bbox[3]
    cdef double x = pt[0]
    cdef double y = pt[1]
    if (xmn <= x < xmx and ymn <= y < ymx):
        return True
    else:
        return False

def hashpt(float xmin, float xmax, float ymin, float ymax, float x, float y):
    """ Returns a generator that returns successive quadrants [0-3] that
    constitute a geohash for *pt* in a global *bbox*. """
    cdef float xm, ym
    cdef int geohash
    while True:
        xm = 0.5 * (xmin + xmax)
        ym = 0.5 * (ymin + ymax)
        if x < xm:
            if y < ym:
                geohash = 0
                xmax = xm
                ymax = ym
            elif y >= ym:
                geohash = 2
                xmax = xm
                ymin = ym
            else:
                raise HashError
        elif x >= xm:
            if y < ym:
                geohash = 1
                xmin = xm
                ymax = ym
            elif y >= ym:
                geohash = 3
                xmin = xm
                ymin = ym
            else:
                raise HashError
        else:
            raise HashError
        yield geohash

class HashError(Exception):
    pass



""" Provides functions for reading and writing ESRI shapefiles and returning a
geometry object. """

import sys
import datetime
import numbers
import shapefile
from .. import errors

try:
    import osgeo
    HAS_OSGEO = True
except ImportError:
    HAS_OSGEO = False

rankerror = IOError("feature must be in two or three dimensions to write "
                    "as shapefile")

def _isnumpyint(o):
    return hasattr(o, "dtype") and \
            o.dtype in ("int8", "int16", "int32", "int64")

def _isnumpyfloat(o):
    return hasattr(o, "dtype") and \
            o.dtype in ("float16", "float32", "float64", "float128")

def _isnumpytype(o):
    return hasattr(o, "dtype")

def property_field_type(value):
    """ Determine the appropriate dBase field type for *value* """
    if isinstance(value, numbers.Number) or _isnumpytype(value):
        ## pyshp doesn't handle the full variety of numeric types
        ## (float, double, long), only 'N'
        #if isinstance(value, numbers.Integral) or _isnumpyint(value):
        #    desc = "I"
        #elif isinstance(value, numbers.Real) or _isnumpyfloat(value):
        #    desc = "O"
        if isinstance(value, (numbers.Integral, numbers.Real)) \
                or _isnumpyint(value) or _isnumpyfloat(value):
            desc = "N"
        else:
            raise TypeError("cannot choose the correct dBase type for "
                            "{0}\n".format(type(value)))
    elif isinstance(value, str):
        desc = "C"
    elif isinstance(value, datetime.datetime):
        desc = "@"
    elif isinstance(value, datetime.date):
        desc = "D"
    elif isinstance(value, bool):
        desc = "L"
    else:
        raise TypeError("cannot choose the correct dBase type for "
                        "{0}\n".format(type(value)))
    return desc

def addfields(writer, properties):
    """ Add geometry properties *properties* to shapefile writer *writer*. """
    writer.field("ID", "I", "8")
    values = []
    for key in properties:
        value = properties[key]
        typ = property_field_type(value)
        dec = 0 if typ not in ("N", "F", "O", "I") else 16
        writer.field(key.upper(), fieldType=typ, size="100", decimal=dec)
        values.append(value)
    writer.record("0", *values)
    return

def addfields_points(writer, points):
    """ Add *points* data to shapefile writer *writer*. """
    writer.field("ID", "I", "8")
    if points.data is not None:
        keys = points.data.fields
        for key in keys:
            t = property_field_type(points.data.get(0)[key])
            dec = 0 if t not in ("N", "F", "O", "I") else 16
            writer.field(key, fieldType=t, size="100", decimal=dec)
        for i, pt in enumerate(points):
            writer.record(str(i), *[pt.data.get(0)[key] for key in keys])
    else:
        for i in range(len(points)):
            writer.record(str(i))
    return

def write_multipoint2(mp, fstem):
    w = shapefile.Writer(shapeType=shapefile.POINT)
    for vertex in mp.vertices:
        w.point(*vertex)
    addfields_points(w, mp)
    w.save(fstem)
    write_projection(mp, fstem)
    return

def write_multipoint3(mp, fstem):
    w = shapefile.Writer(shapeType=shapefile.POINTZ)
    for vertex in mp.vertices:
        w.point(*vertex)
    addfields_points(w, mp)
    w.save(fstem)
    write_projection(mp, fstem)
    return

def write_line2(line, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYLINE)
    w.poly(shapeType=shapefile.POLYLINE, parts=[line.vertices])
    addfields(w, line.properties)
    w.save(fstem)
    write_projection(line, fstem)
    return

def write_line3(line, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYLINEZ)
    w.poly(shapeType=shapefile.POLYLINEZ, parts=[line.vertices])
    addfields(w, line.properties)
    w.save(fstem)
    write_projection(line, fstem)
    return

def write_poly2(poly, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYGON)
    w.poly(shapeType=shapefile.POLYGON, parts=[poly.vertices])
    addfields(w, poly.properties)
    w.save(fstem)
    write_projection(poly, fstem)
    return

def write_poly3(poly, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYGONZ)
    w.poly(shapeType=shapefile.POLYGONZ, parts=[poly.vertices])
    addfields(w, poly.properties)
    w.save(fstem)
    write_projection(poly, fstem)
    return

def shapefile_type(feature):
    """ Match a feature with a shapefile type identifier """
    if feature._geotype == "Multipoint":
        if feature.rank == 2:
            return shapefile.POINT
        elif feature.rank == 3:
            return shapefile.POINTZ
        else:
            raise rankerror
    elif feature._geotype == "Line":
        if feature.rank == 2:
            return shapefile.POLYLINE
        elif feature.rank == 3:
            return shapefile.POLYLINEZ
        else:
            raise rankerror
    elif feature._geotype == "Polygon":
        if feature.rank == 2:
            return shapefile.POLYGON
        elif feature.rank == 3:
            return shapefile.POLYGONZ
        else:
            raise rankerror
    else:
        raise TypeError("Expected list with items of geotype 'Multipoint',"
                        "'Line', or 'Polygon'")

def write_shapefile(features, fstem):
    """ Write *features* to a shapefile. All features must be of the same type
    (i.e. Multipoints, Lines, Polygons). """
    if len(features) == 1:
        features[0].to_shapefile(fstem)
        return
    elif not all(f._geotype == features[0]._geotype for f in features[1:]):
        raise IOError("all features must be of the same type")
    elif not all(f.rank == features[0].rank for f in features[1:]):
        raise IOError("all features must have the same dimensionality")

    shapeType = shapefile_type(features[0])
    w = shapefile.Writer(shapeType=shapeType)
    if shapeType in (shapefile.POINT, shapefile.POINTZ):

        # add geometry
        for feature in features:
            for pt in feature:
                w.point(*pt.vertex)

        # add records
        w.field("ID", "I", "8")
        if features[0].data is not None:
            keys = set(features[0].data.fields)     # for testing similarity
            keylist = features[0].data.fields       # preserves order
        else:
            keys = []

        if len(keys) != 0 and all(keys == set(f.data.fields) for f in features[1:]):
            for key in keylist:
                testvalue = features[0].data[key][0]
                w.field(key.upper(), property_field_type(testvalue), "100")
            i = 0
            for feature in features:
                for pt in feature:
                    w.record(str(i), *[pt.data[key] for key in keylist])
                    i += 1
        else:
            i = 0
            for feature in features:
                for pt in feature:
                    w.record(str(i))
                    i += 1

    else:

        # add geometry
        for feature in features:
            w.poly([feature.vertices])

        # add records
        w.field("ID", "I", "8")
        keys = set(features[0].properties.keys())   # for testing similarity
        keylist = list(keys)                        # preserves order

        if len(keys) != 0 and all(keys == set(f.properties.keys()) for f in features[1:]):
            for key in features[0].properties:
                value = features[0].properties[key]
                propertyType = property_field_type(value)
                length = "100"
                w.field(key.upper(), propertyType, length)
            for (i, feature) in enumerate(features):
                values = []
                for key in keylist:
                    values.append(feature.properties[key])
                w.record(str(i), *values)
        else:
            for (i, feature) in enumerate(features):
                w.record(str(i))

    w.save(fstem)
    write_projection(features[0], fstem)
    return

def write_projection(geom, fstem):
    try:
        wkt = geom.crs.get_wkt()
        with open(fstem+".prj", "w") as f:
            f.write(wkt)
    except errors.CRSError:
        sys.stderr.write("Projection information not saved (requires osgeo)\n")
    except AttributeError as e:
        sys.stderr.write(str(e))
        sys.stderr.write("\n")




"""
Geographical measurement and simple analysis module for Python. Provides
Point, Multipoint, Line, and Polygon classes, with methods for simple
measurements such as distance, area, and bearing.
"""
from __future__ import division

import math
import sys
import itertools
import warnings
import numpy as np
from . import vtk
from . import geojson
from . import xyfile
from . import shp
from .metadata import Metadata, Indexer
from . import quadtree
from . import _cvectorgeo
from .. import geodesy
from ..crs import Cartesian, GeographicalCRS, SphericalCRS, EllipsoidalCRS
from ..crs import SphericalEarth
from ..errors import GeometryError, GGeoError, GUnitError, GInitError, CRSError

class Geometry(object):
    """ This is the abstract base class for all geometry types """
    _geotype = None

    def __init__(self, crs=Cartesian):
        self.properties = {}
        self.crs = crs
        return

    @staticmethod
    def _distance(pos0, pos1):
        """ Generic method for calculating distance between positions that
        respects CRS """
        if pos0.crs == pos1.crs:
            (x0,y0) = pos0.x, pos0.y
            (x1,y1) = pos1.x, pos1.y
            ((x0,x1), (y0,y1)) = pos0.crs.project([x0, x1], [y0, y1],
                                                  inverse=True)
            _, _, dist = pos0.crs.inverse(x0, y0, x1, y1, radians=False)
        else:
            raise CRSError("Positions must use the same CRS")
        return dist

    @property
    def d(self):
        return Indexer(self.data)

    def add_property(self, name, value):
        """ Insert a property (name -> value) into the properties dict, raising
        a NameError if the property already exists. Compared to directly
        manipulating the properties dict, this avoids accidents. """
        if name not in self.properties:
            self.properties[name] = value
        else:
            raise NameError("property '{0}' already exists in {1}".format(name, type(self)))
        return

    def delete_property(self, name):
        """ Equivalent to

            del self.properties[name]
        """
        if name in self.properties:
            del self.properties[name]
        return


class Point(Geometry):
    """ Point object instantiated with:

    *coords*        2-tuple or 3-tuple
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Point"

    def __init__(self, coords, data=None, properties=None, copy_metadata=False,
                 **kwargs):
        try:
            self.rank = 2 if len(coords) == 2 else 3
        except TypeError:
            raise TypeError("Point coordinates must be a sequence")
        super(Point, self).__init__(**kwargs)
        self.vertex = coords

        if data is None:
            self.data = data
        else:
            self.data = Metadata(data)

        if hasattr(properties, "keys"):
            self.properties = properties
        else:
            properties = {}
        return

    def __getitem__(self, idx):
        return self.vertex[idx]

    def __repr__(self):
        return 'Point(' + str(self.vertex) + ')'

    def __eq__(self, other):
        try:
            return (tuple(self.vertex) == tuple(other.vertex)) and \
                   (self.data == other.data) and \
                   (self.properties == other.properties) and \
                   (self.crs == other.crs)
        except AttributeError:
            return False

    def __neq__(self, other):
        return ~(self == other)

    @property
    def __geo_interface__(self):
        return {"type" : "Point", "coordinates" : self.vertex}

    @property
    def x(self):
        return self.vertex[0]

    @property
    def y(self):
        return self.vertex[1]

    @property
    def z(self):
        return self.vertex[2]

    def get_vertex(self, crs=None):
        """ Return the Point vertex as a tuple. """
        if crs is None or crs==self.crs:
            return self.vertex
        else:
            vg = self.crs.project(self.x, self.y, inverse=True)
            return crs.project(*vg)

    def coordsxy(self, convert_to=False):
        """ Returns the x,y coordinates. Convert_to may be set to 'deg'
        or 'rad' for convenience.  """
        if convert_to == 'rad':
            return (self.x*3.14159/180., self.y*3.14159/180.)
        elif convert_to == 'deg':
            return (self.x/3.14159*180., self.y/3.14159*180.)
        else:
            return (self.x, self.y)

    def azimuth(self, other, crs=SphericalEarth):
        """ Returns the compass azimuth from self to other in radians (i.e.
        clockwise, with north at 0). Returns NaN if points are coincident. """

        if (self.x, self.y) == (other.x, other.y):
            az = np.nan

        elif self.crs == other.crs:
            lon0, lat0 = self.crs.project(self.x, self.y, inverse=True)
            lon1, lat1 = self.crs.project(other.x, other.y, inverse=True)
            az, _, _ = self.crs.inverse(lon0, lat0, lon1, lat1)

        else:
            raise CRSError("Azimuth undefined for points in CRS {0} and "
                           "{1}".format(self.crs, other.crs))
        return az

    def walk(self, distance, direction, radians=False):
        """ Returns the point reached when moving in a given direction for
        a given distance from a specified starting location.

            distance (float): distance to walk
            direction (float): walk azimuth (clockwise with "north" at 0)
        """
        xg1, yg1 = self.crs.project(self.x, self.y, inverse=True)
        xg2, yg2, _ = self.crs.forward(xg1, yg1, direction, distance,
                                        radians=radians)
        x, y = self.crs.project(xg2, yg2)
        return Point((x, y), properties=self.properties, data=self.data,
                     crs=self.crs, copy_metadata=False)

    def distance(self, other):
        """ Returns a distance to another Point. If the coordinate system is
        geographical and a third (z) coordinate exists, it is assumed to have
        the same units as the real-world horizontal distance (i.e. meters). """
        if self.crs != other.crs:
            raise CRSError("Points must share the same coordinate system.")
        flat_dist = self._distance(self, other)
        if 2 == self.rank == other.rank:
            return flat_dist
        else:
            return math.sqrt(flat_dist**2. + (self.z-other.z)**2.)

    def shift(self, shift_vector):
        """ Shift point by the amount given by a vector. Operation occurs
        in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        self.vertex = tuple([a+b for a,b in zip(self.vertex, shift_vector)])
        return self

    def as_geojson(self, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer


class MultipointBase(Geometry):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes. """
    _geotype = "MultipointBase"

    def __init__(self, vertices, data=None, properties=None, copy_metadata=False,
                 **kwargs):
        """ Partial init function that establishes geometry rank and creates a
        metadata attribute.
        """
        super(MultipointBase, self).__init__(**kwargs)
        vertices = list(vertices)
        if len(vertices) > 0:

            def ispoint(a):
                return getattr(a, "_geotype", None) == "Point"

            if all(ispoint(a) for a in vertices):
                pts = vertices

                # Consolidate a list of points
                rank = pts[0].rank
                crs = pts[0].crs
                if len(pts) != 1 and any(crs != pt.crs for pt in pts[1:]):
                    raise CRSError("All points must share the same CRS")
                if len(pts) != 1 and any(rank != pt.rank for pt in pts[1:]):
                    raise GInitError("Input must have consistent rank")

                # Data
                if pts[0].data is not None:
                    if all(pt.data._fields == pts[0].data._fields for pt in pts[1:]):
                        if data is not None:
                            raise GInitError("Data kweyword disallowed when constructing from points")
                        d = [pt.data._data[0] for pt in pts]
                        self.data = Metadata(d, fields=pts[0].data.fields)
                    else:
                        raise GInitError("Point have inconsistent data attributes")
                elif data is not None:
                    self.data = Metadata(data)
                else:
                    self.data = None

                self.vertices = [pt.vertex for pt in pts]
                self.rank = rank
                self.crs = crs

            else:

                # Construct from a list of positions (tuples)

                self.rank = len(vertices[0])

                if not 2 <= self.rank <= 3:
                    raise GInitError("Input must be doubles or triples")
                if any(self.rank != len(v) for v in vertices):
                    raise GInitError("Input must have consistent rank")

                self.vertices = [tuple(v) for v in vertices]

                if data is None:
                    self.data = None
                else:
                    self.data = Metadata(data)

        else:
            self.rank = None
            self.vertices = []

        if hasattr(properties, "keys"):
            self.properties = properties
        elif properties is None:
            self.properties = {}
        else:
            raise GInitError("value provided as 'properties' must be a hash")

        self.quadtree = None
        return

    def __repr__(self):
        if len(self) < 5:
            ppverts = str(self.vertices)
        else:
            ppverts = str(self.vertices[:2])[:-1] + "..." + str(self.vertices[-2:])[1:]
        return '{typ}({verts})>'.format(
                typ=str(type(self))[:-1], verts=ppverts)

    def __len__(self):
        return len(self.vertices)

    def __getitem__(self, key):
        d = None
        if isinstance(key, (int, np.int64)):
            if self.data is not None:
                d = Metadata([self.data[key]], fields=self.data._fields)
            return Point(self.vertices[key], data=d, properties=self.properties,
                         crs=self.crs)
        elif isinstance(key, slice):
            if self.data is not None:
                d = Metadata(self.data[key], fields=self.data._fields)
            return type(self)(self.vertices[key], data=d,
                              properties=self.properties, crs=self.crs)
        else:
            raise GGeoError('Index must be an integer or a slice object')

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise GGeoError('Indices must be integers')
        try:
            self.vertices[key] = value.vertex
            if None not in (self.data, value.data):
                self.data[key] = tuple(value.data[value.data._fields.index(f)] for f in self._fields)
            elif self.data is not None:
                self.data[key] = (None for i in range(len(self.data._fields)))

        except AttributeError:
            self.vertices[key] = value
            if self.data is not None:
                self.data[key] = (None for i in range(len(self.data._fields)))

        if getattr(value, "_geotype", None) == "Point":
            self.vertices[key] = value.vertex
            if self.data is not None:
                self.data[key] = value.data._data[0]
        elif len(value) == self.rank:
            self.vertices[key] = value
        else:
            raise GGeoError('Cannot insert non-Pointlike value with '
                            'length != {0}'.format(self.rank))
        return

    def __delitem__(self, key):
        if len(self) > key:
            del self.vertices[key]
            del self.data[key]
        else:
            raise GGeoError('Index ({0}) exceeds length'
                            '({1})'.format(key, len(self)))
        return

    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __eq__(self, other):
        try:
            return (self._geotype == other._geotype) and \
                   (self.vertices == other.vertices) and \
                   (self.data == other.data) and \
                   (self.properties == other.properties) and \
                   (self.crs == other.crs)
        except (AttributeError, TypeError):
            return False

    def __neq__(self, other):
        return ~(self == other)

    def _bbox_overlap(self, other):
        """ Return whether bounding boxes between self and another geometry
        overlap.
        """
        reg0 = self.bbox
        reg1 = other.bbox
        return (reg0[0] <= reg1[2] and reg1[0] <= reg0[2] and
                reg0[1] <= reg1[3] and reg1[1] <= reg0[3])

    @property
    def bbox(self):
        """ Return the extent of a bounding box as
            (xmin, ymin, xmax, ymax)
        """
        x, y = self.get_coordinate_lists()
        bbox = (min(x), min(y), max(x), max(y))
        return bbox

    @property
    def coordinates(self):
        return self.get_coordinate_lists()

    def print_vertices(self):
        """ Prints an enumerated list of indices. """
        for i, vertex in enumerate(self.vertices):
            print("{0}\t{1}".format(i, vertex))

    def get_vertices(self, crs=None):
        """ Return vertices as an array. """
        if (crs is None) or (crs is self.crs):
            return np.array(self.vertices)
        else:
            vertices = [crs.project(*self.crs.project(*v[:2], inverse=True))
                        for v in self.vertices]
            return np.array(vertices)

    def get_coordinate_lists(self, crs=None):
        """ Return horizontal coordinate lists, optionally projected to *crs*.
        """
        if self.rank == 2:
            x, y = tuple(zip(*self.vertices))
        else:
            x, y, _ = tuple(zip(*self.vertices))
        if crs is not None and (crs != self.crs):
            xg, yg = self.crs.project(x, y, inverse=True)
            x, y = crs.project(xg, yg)
        return x, y

    def shift(self, shift_vector):
        """ Shift feature by the amount given by a vector. Operation
        occurs in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        if self.rank == 2:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1])
        elif self.rank == 3:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1],
                            pt[2] + shift_vector[2])
        self.vertices = list(map(f, self.vertices))
        self.quadtree = None
        return self

    def _matmult(self, A, x):
        """ Return Ax=b """
        b = []
        for a in A:
            b.append(sum([ai * xi for ai, xi in zip(a, x)]))
        return b

    def rotate2d(self, thetad, origin=(0, 0)):
        """ Rotate rank 2 Multipoint around *origin* counter-clockwise by
        *thetad* degrees. """
        # First, shift by the origin
        self.shift([-a for a in origin])

        # Multiply by a rotation matrix
        theta = thetad / 180.0 * math.pi
        R = ((math.cos(theta), -math.sin(theta)),
             (math.sin(theta), math.cos(theta)))
        rvertices = [self._matmult(R, x) for x in self.vertices]
        self.vertices = rvertices

        # Shift back
        self.shift(origin)
        self.quadtree = None
        return self

    def apply_affine_transform(self, M):
        """ Apply an affine transform given by matrix *M* to data and return a
        new geometry. """
        vertices = []
        for x,y in self.get_vertices():
            vertices.append(tuple(np.dot(M, [x, y, 1])[:2]))
        return type(self)(vertices, data=self.data, properties=self.properties,
                          crs=self.crs)

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        if self.data is not None:
            data = Metadata([self.data[i] for i in idxs], fields=self.data._fields)
        else:
            data = None
        subset = type(self)(vertices, data=data, properties=self.properties,
                            crs=self.crs, copy_metadata=False)
        return subset

    def flat_distances_to(self, pt):
        """ Return the "flat Earth" distance from each vertex to a point. """
        A = np.array(self.vertices)
        P = np.tile(np.array(pt.vertex), (A.shape[0], 1))
        d = np.sqrt(np.sum((A-P)**2, 1))
        return d

    def distances_to(self, pt):
        """ Return the distance from each vertex to a point. """
        d = [pt.distance(a) for a in self]
        return np.array(d)

    def nearest_point_to(self, pt):
        """ Returns the internal Point that is nearest to *pt*. If two points
        are equidistant, only one will be returned.
        """
        distances = self.distances_to(pt)
        idx = np.argmin(distances)
        return self[idx]

    def get_extent(self, crs=None):
        """ Calculate a bounding box. """
        def gen_minmax(G):
            """ Get the min/max from a single pass through a generator. """
            (xmin, ymin) = (xmax, ymax) = next(G)
            for (x, y) in G:
                xmin = min(xmin, x)
                xmax = max(xmax, x)
                ymin = min(ymin, y)
                ymax = max(ymax, y)
            return xmin, xmax, ymin, ymax

        if (crs is None) or (crs == self.crs):
            xmin, xmax, ymin, ymax = gen_minmax(c[:2] for c in self.vertices)
        else:
            fprj = lambda c: crs.project(*self.crs.project(*c, inverse=True))
            xmin, xmax, ymin, ymax = gen_minmax(fprj(c[:2]) for c in self.vertices)
        return xmin, xmax, ymin, ymax

    def get_extents(self, crs=None):
        warnings.warn("method `get_extents` has been renamed `get_extent`",
                FutureWarning)
        return self.get_extent(crs=crs)

    def any_within_poly(self, poly):
        """ Return whether any vertices are inside *poly* """
        for pt in self:
            if poly.contains(pt):
                return True
        return False

    def convex_hull(self):
        """ Return a Polygon representing the convex hull. Assumes that the CRS
        may be treated as cartesian. """
        points = [pt for pt in self]

        # Find the lowermost (left?) point
        pt0 = points[0]
        idx = 0
        for i, pt in enumerate(points[1:]):
            if (pt.y < pt0.y) or ((pt.y == pt0.y) and (pt.x < pt0.x)):
                pt0 = pt
                idx = i+1
        points.pop(idx)

        # Sort CCW relative to pt0, and drop all but farthest of any duplicates
        points.sort(key=lambda pt: pt0.distance(pt))
        points.sort(key=lambda pt: _cvectorgeo.polarangle(pt0.vertex, pt.vertex))
        alpha = -1
        drop = []
        for i,pt in enumerate(points):
            a = _cvectorgeo.polarangle(pt0.vertex, pt.vertex)
            if a == alpha:
                drop.append(i)
            else:
                alpha = a

        if len(drop) != 0:
            for i in drop[::-1]:
                points.pop(i)

        # initialize convex hull
        if len(points) == 2:
            return [pt0, points[0], points[1]]
        elif len(points) == 1:
            return [pt0, points[0]]
        else:

            S = [pt0, points[0], points[1]]
            for pt in points[2:]:
                while not _cvectorgeo.isleft(S[-2].vertex, S[-1].vertex, pt.vertex):
                    S.pop()
                S.append(pt)

        return Polygon(S, crs=self.crs)

    def to_xyfile(self, fnm, fields=None, delimiter=' ', header=None):
        """ Write data to a delimited ASCII table.

        fnm         :   filename to write to

        kwargs:
        fields      :   specify the fields to be written (default all)

        Additional kwargs are passed to `xyfile.write_xy`.
        """
        xyfile.write_xy(self.get_vertices(), fnm, delimiter=delimiter, header=header)
        return

    def as_geojson(self, **kwargs):
        """ Print representation of internal data as a GeoJSON string.

        Parameters
        ----------
        crs : coordinate reference system
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, crs=self.crs, **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, crs=self.crs, **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer

    def to_vtk(self, f, **kwargs):
        """ Write data to an ASCII VTK .vtp file. """
        if not hasattr(f, "write"):
            with open(f, "w") as fobj:
                vtk.mp2vtp(self, fobj, **kwargs)
        else:
            vtk.mp2vtp(self, f, **kwargs)
        return

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            shp.write_multipoint2(self, fstem)
        elif self.rank == 3:
            shp.write_multipoint3(self, fstem)
        else:
            raise IOError("Rank must be 2 or 3 to write as a shapefile")
        return


class Multipoint(MultipointBase):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Multipoint"

    def __contains__(self, other):
        if other in (pt for pt in self):
            return True
        else:
            return False

    @property
    def __geo_interface__(self):
        return {"type" : "MultiPoint", "bbox" : self.bbox, "coordinates" : self.vertices}

    def within_radius(self, pt, radius):
        """ Return Multipoint of subset that is within *radius* of *pt*.
        """
        if self.quadtree is None:
            distances = self.distances_to(pt)
            indices = [i for i,d in enumerate(distances) if d <= radius]
        else:
            search_bbox = (pt.x-radius, pt.x+radius, pt.y-radius, pt.y+radius)
            possible_pt_tuples = self.quadtree.getfrombbox(search_bbox)
            possible_pts = []
            for t in possible_pt_tuples:
                possible_pts.append(Point((t[0], t[1]), properties={"idx": t[2]},
                                                        crs=self.crs))
            indices = [p.properties["idx"] for p in possible_pts
                            if pt.distance(p) <= radius]

        return self._subset(indices)

    def within_bbox(self, bbox):
        """ Return Multipoint subset that is within a square bounding box
        given by (xmin, xymin, xmax, ymax).
        """
        filtbbox = lambda pt: (bbox[0] <= pt.vertex[0] <= bbox[2]) and \
                              (bbox[1] <= pt.vertex[1] <= bbox[3])
        indices = [i for (i, pt) in enumerate(self) if filtbbox(pt)]
        return self._subset(indices)

    def within_polygon(self, poly):
        """ Return Multipoint subset that is within a polygon.
        """
        if self.quadtree is None:
            indices = [i for (i, pt) in enumerate(self) if poly.contains(pt)]
        else:
            search_bbox = poly.get_extent(crs=self.crs)
            possible_pt_tuples = self.quadtree.getfrombbox(search_bbox)
            possible_pts = []
            for t in possible_pt_tuples:
                possible_pts.append(Point((t[0], t[1]), properties={"idx": t[2]},
                                                        crs=self.crs))
            indices = [p.properties["idx"] for p in possible_pts
                            if poly.contains(p)]
        return self._subset(indices)

    def build_quadtree(self, buffer=1e-8):
        """ Construct an internal quadtree with the current geometry data.

        Optional *buffer* keyword specifies a spatial buffer to create around
        the current point bounding box, permitting the geometry to grow after
        the quadtree has been initialized. *buffer* may be a scalar or a
        sequence of (left, right, bottom, top).
        """
        try:
            bf = (buffer[0], buffer[1], buffer[2], buffer[3])
        except TypeError:
            bf = (buffer, buffer, buffer, buffer)

        x0, x1, y0, y1 = self.get_extent()
        self.quadtree = quadtree.QuadTree((x0-bf[0], x1+bf[1], y0-bf[2], y1+bf[3]))
        for i, pt in enumerate(self):
            self.quadtree.addpt((pt.x, pt.y, i))
        return

class ConnectedMultipoint(MultipointBase):
    """ Class for Multipoints in which vertices are assumed to be connected. """

    @property
    def bbox(self):
        if isinstance(self.crs, GeographicalCRS):
            xmin = xmax = self[0].x
            ymin = ymax = self[0].y
            rot = 0
            for seg in self.segments:
                x0, x1 = seg[0].x, seg[1].x
                ymin = min(ymin, seg[1].y)
                ymax = max(ymax, seg[1].y)
                if self._seg_crosses_dateline(seg):
                    if x0 > x1:     # east to west
                        rot += 360
                    else:           # west to east
                        rot -= 360

                    xmin = min(xmin, x1+rot)
                    xmax = max(xmax, x1+rot)

            xmin = (xmin+180) % 360 - 180
            xmax = (xmax+180) % 360 - 180
            return (xmin, ymin, xmax, ymax)
        else:
            return super(ConnectedMultipoint, self).bbox

    @property
    def length(self):
        """ Returns the length of the line/boundary. """
        points = [Point(v, crs=self.crs) for v in self.vertices]
        distances = [a.distance(b) for a, b in zip(points[:-1], points[1:])]
        return sum(distances)

    @property
    def segments(self):
        """ Returns an generator of adjacent line segments. """
        return (self._subset((i,i+1)) for i in range(len(self)-1))

    @property
    def segment_tuples(self):
        """ Returns an generator of adjacent line segments as coordinate tuples. """
        return ((self.vertices[i], self.vertices[i+1])
                for i in range(len(self.vertices)-1))

    def intersects(self, other):
        """ Return whether an intersection exists with another geometry. """
        interxbool = (np.nan in _cvectorgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                                     a[0][1], a[1][1], b[0][1], b[1][1])
                    for a in self.segments for b in other.segments)
        if self._bbox_overlap(other) and (True not in interxbool):
            return True
        else:
            return False

    def intersections(self, other, keep_duplicates=False):
        """ Return the intersections with another geometry as a Multipoint. """
        interx = (_cvectorgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                          a[0][1], a[1][1], b[0][1], b[1][1])
                     for a in self.segments for b in other.segments)
        if not keep_duplicates:
            interx = set(interx)
        interx_points = []
        for vertex in interx:
            if np.nan not in vertex:
                interx_points.append(Point(vertex, properties=self.properties,
                                           crs=self.crs))
        return Multipoint(interx_points)

    def _nearest_to_point(self, pt):
        """ Return a tuple of the shortest distance on the geometry boundary to
        *pt*, and the vertex at that location. """
        ptvertex = pt.get_vertex(crs=self.crs)
        segments = zip(self.vertices[:-1], self.vertices[1:])

        if self.crs == Cartesian:
            func = _cvectorgeo.pt_nearest_planar
            func = lambda ptseg: _cvectorgeo.pt_nearest_planar(ptseg[0],
                                            ptseg[1][0], ptseg[1][1])
        else:
            fwd = self.crs.forward
            inv = self.crs.inverse
            func = lambda ptseg: _cvectorgeo.pt_nearest_proj(fwd, inv, ptseg[0],
                                            ptseg[1][0], ptseg[1][1], tol=0.01)

        point_dist = map(func, zip(itertools.repeat(ptvertex), segments))
        minpt = None
        mindist = -1.0
        for i, (pt, d) in enumerate(point_dist):
            if d < mindist or (i == 0):
                minpt = pt
                mindist = d
        return mindist, minpt

    def shortest_distance_to(self, pt):
        """ Return the shortest distance from any position on the geometry
        boundary to *pt* (Point).
        """
        return self._nearest_to_point(pt)[0]

    def nearest_on_boundary(self, pt):
        """ Returns the position on the geometry boundary that is nearest to
        *pt* (Point). If two points are equidistant, only one will be returned.
        """
        _, minpt = self._nearest_to_point(pt)
        return Point(minpt, crs=self.crs)

    def within_distance(self, pt, distance):
        """ Test whether a point is within *distance* of a ConnectedMultipoint.
        """
        return all(distance >= seg.shortest_distance_to(pt) for seg in self.segments)

    @staticmethod
    def _seg_crosses_dateline(seg):
        a, b = seg[0], seg[1]
        return (sign(a.x) != sign(b.x)) and (abs(a.x-b.x) > 180.0)

    def crosses_dateline(self):
        """ Return a boolean that indicates whether any segment crosses the
        dateline """
        if not isinstance(self.crs, GeographicalCRS):
            raise CRSError("Dateline detection only defined for geographical "
                           "coordinates")

        return any(self._seg_crosses_dateline(seg) for seg in self.segments)


class Line(ConnectedMultipoint):
    """ Line composed of connected vertices.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Line"

    @property
    def __geo_interface__(self):
        return {"type" : "LineString", "bbox" : self.bbox, "coordinates" : self.vertices}

    def extend(self, other):
        """ Combine two lines, provided that that the data formats are similar.
        """
        if self.rank != other.rank:
            raise GGeoError("Rank mismatch ({0} != {1})".format(self.rank, other.rank))
        if self._geotype != other._geotype:
            raise GGeoError("Geometry mismatch ({0} != {1})".format(self._geotype, other._geotype))

        if None not in (self.data, other.data):
            self.data._data.extend(other.data._data)
        elif self.data == other.data:
            self.data = None
        else:
            raise GGeoError('Cannot add geometries with mismatched metadata')
        self.vertices.extend(other.vertices)
        self.quadtree = None
        return self

    def cumulength(self):
        """ Returns the cumulative length by vertex. """
        d = [0.0]
        pta = self[0]
        for ptb in self[1:]:
            d_ = pta.distance(ptb)
            d.append(d_ + d[-1])
            pta = ptb
        return d

    def subsection(self, n):
        """ Return *n* equally spaced Point instances along line. """
        segments = self.segments
        Ltotal = self.cumulength()[-1]
        step = Ltotal / float(n-1)
        step_remaining = step

        points = [self[0]]
        x = 0.0
        pos = self[0]
        seg = next(segments)
        seg_remaining = seg.displacement()

        while x < Ltotal-1e-8:
            direction = seg[0].azimuth(seg[1])

            if step_remaining <= seg_remaining:
                pos = pos.walk(step_remaining, direction)
                x += step_remaining
                seg_remaining -= step_remaining
                step_remaining = step
                points.append(pos)
                seg.vertices[0] = pos.vertex

            else:
                pos = seg[1]
                x += seg_remaining
                step_remaining -= seg_remaining

                seg = next(segments, seg)
                seg_remaining = seg.displacement()
                # except StopIteration as e:
                #     if abs(Ltotal-x) > 1e-8:       # tolerance for endpoint
                #         raise e

        if len(points) == n-1:
            points.append(seg[-1])
        return points

    def displacement(self):
        """ Returns the distance between the first and last vertex. """
        return self[0].distance(self[-1])

    def to_polygon(self):
        """ Returns a polygon. """
        return Polygon(self.vertices, data=self.data, properties=self.properties,
                       crs=self.crs, copy_metadata=True)

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            shp.write_line2(self, fstem)
        elif self.rank == 3:
            shp.write_line3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile")
        return


class Polygon(ConnectedMultipoint):
    """ Polygon, composed of a closed sequence of vertices.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *subs*          List of sub-polygons [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Polygon"
    subs = []

    def __init__(self, vertices, data=None, properties=None, subs=None, **kwargs):
        vertices = list(vertices)
        ConnectedMultipoint.__init__(self, vertices, data=data,
                                     properties=properties, **kwargs)
        self.subs = subs if subs is not None else []
        return

    def __getitem__(self, key):
        if isinstance(key, slice):
            ind = key.indices(len(self))
            if len(self) != ((ind[1] - ind[0]) // ind[2]):
                if self.data is None:
                    d = None
                else:
                    d = self.data[key]
                return Line(self.vertices[key], data=d,
                            properties=self.properties, crs=self.crs)
        return super(Polygon, self).__getitem__(key)

    @property
    def __geo_interface__(self):
        coords = [self.vertices]
        for geom in self.subs:
            coords.append(geom.vertices)
        return {"type" : "Polygon", "bbox" : self.bbox, "coordinates" : coords}

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        if self.data is None:
            data = None
        else:
            data = Metadata([self.data[i] for i in idxs], fields=self.data.fields)
        subset = Line(vertices, data=data, properties=self.properties,
                      crs=self.crs, copy_metadata=False)
        return subset

    def isclockwise(self):
        """ Return whether polygon winds clockwise around its interior. """
        s = sum((seg[1][0] - seg[0][0]) * (seg[1][1] + seg[0][1])
                for seg in self.segments)
        return s > 0

    def ispolar(self, pole=None):
        """ Return True if polygon contains one pole. If the polygon contains
        neither or both poles, returns False.

        Currently expects geographical coordinates with longitude ranging from
        (-180, 180].
        """

        if not isinstance(self.crs, GeographicalCRS):
            raise CRSError("ispolar defined only for geographical CRS")

        if pole is None:
            pole = Point((0, 90), crs=SphericalEarth)

        lon0 = self[-1].vertex[0]
        sum_angle = 0.0

        for vertex in self.vertices:

            lon1 = vertex[0]

            if (sign(lon0) == -sign(lon1)) and \
                ((min(abs(lon0-180), abs(lon0+180)) + min(abs(lon1-180), abs(lon1+180))) < (abs(lon0) + abs(lon1))):
                # Longitudes span the dateline
                sum_angle += 360.0 + lon1 - lon0
            else:
                sum_angle += lon1 - lon0

            lon0 = lon1

        return True if abs(sum_angle) > 1e-4 else False

    @property
    def segments(self):
        """ Returns an generator of adjacent line segments.
        Unique to Polygon: appends a final segment to close the Polygon.
        """
        L = len(self.vertices)
        return itertools.chain((self._subset((i,i+1)) for i in range(len(self)-1)),
                               (self._subset((L-1,0)),))

    @property
    def segment_tuples(self):
        """ Returns an generator of adjacent line segments as coordinate tuples. """
        return ((self.vertices[i-1], self.vertices[i]) for i in range(len(self.vertices)))

    @property
    def length(self):
        raise AttributeError("%s instance has no attribute 'length'" % type(self))

    @property
    def perimeter(self):
        """ Return the perimeter of the polygon. If there are sub-polygons,
        their perimeters are added recursively. """
        return sum(seg.length for seg in self.segments) + \
                sum([p.perimeter for p in self.subs])

    @property
    def area(self):
        """ Return the two-dimensional area of the polygon, excluding
        sub-polygons. """

        if isinstance(self.crs, EllipsoidalCRS):
            a = 0.0
            for seg in self.segment_tuples:
                x1, y1 = seg[0]
                x2, y2 = seg[1]
                a_, b_ = self.crs.a, self.crs.b
                a+= geodesy.ellipsoidal_area(a_, b_, x1, y1, x2, y2)

        elif isinstance(self.crs, SphericalCRS):
            a = 0.0
            for seg in self.segment_tuples:
                x1, y1 = seg[0]
                x2, y2 = seg[1]
                a += geodesy.spherical_area(self.crs.radius, x1, y1, x2, y2)

        elif isinstance(self.crs, GeographicalCRS):
            # Catch anything that's not spherical or ellipsoidal
            raise CRSError("area not implemented for unknown geographical CRS\n"
                           "try replacing the CRS with a known SphericalCRS or "
                           "EllipsoidalCRS class")

        else:
            # All projected coordinate systems
            x, y = self.coordinates
            x0 = np.min(x)
            a = (0.5*(x[0] + x[-1]) - x0) * (y[0] - y[-1])
            a += sum((0.5*(x[i+1]+x[i]) - x0) * (y[i+1] - y[i]) for i in range(len(x)-1))
        return abs(a) - sum(sub.area for sub in self.subs)

    @property
    def centroid(self):
        """ Return Polygon centroid as a Point, ignoring sub-polygons. """
        x, y = self.coordinates
        A = 0.5 * sum(x[i]*y[i+1] - x[i+1]*y[i] for i in range(-1, len(self)-1))
        cx = sum((x[i] + x[i+1]) * (x[i]*y[i+1] - x[i+1]*y[i])
                    for i in range(-1, len(self)-1)) / (6*A)
        cy = sum((y[i] + y[i+1]) * (x[i]*y[i+1] - x[i+1]*y[i])
                    for i in range(-1, len(self)-1)) / (6*A)
        return Point((cx, cy), properties=self.properties, crs=self.crs)

    @staticmethod
    def _signcross(a, b):
        """ Return sign of 2D cross product a x b """
        c = (a[0]*b[1]) - (a[1]*b[0])
        if c != 0:
            return c/abs(c)
        else:
            return 0

    def contains(self, pt):
        """ Returns True if pt is inside or on the boundary of the polygon, and
        False otherwise. Uses a crossing number scheme.

        Behaviour may not be defined for polar geographical polygons.
        """
        if isinstance(self.crs, GeographicalCRS):
            if self.ispolar():
                raise NotImplementedError(
                    "Membership tests on polar geographical polygons not "
                    "implemented. As a workaround, transform to an appropriate "
                    "ProjectedCRS first.")

        x, y = pt.get_vertex(crs=self.crs)[:2]
        cnt = 0
        for seg in self.segment_tuples:
            (a, b) = seg
            if _cvectorgeo.intersects_cn(x, y, a[0], b[0], a[1], b[1]):
                cnt += 1
        return cnt % 2 == 1 and not any(p.contains(pt) for p in self.subs)

    def to_line(self):
        """ Returns a self-closing polyline. Discards sub-polygons. """
        v = self.vertices + self.vertices[0]
        return Line(v, properties=self.properties, data=self.data,
                    crs=self.crs, copy_metadata=True)

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            shp.write_poly2(self, fstem)
        elif self.rank == 3:
            shp.write_poly3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile")
        return


def points_to_multipoint(points):
    """ Merge *points* into a Multipoint instance. Point properties are stored
    as Multipoint data. All points must use the same CRS.
    """
    crs = points[0].crs
    if not all(pt.crs == crs for pt in points):
        raise CRSError("All points must share the same CRS")

    keys = list(points[0].properties.keys())
    for pt in points[1:]:
        for key in keys:
            if key not in pt.properties:
                keys.pop(keys.index(key))

    ptdata = {}
    for key in keys:
        ptdata[key] = [pt.properties[key] for pt in points]

    vertices = [pt.vertex for pt in points]

    return Multipoint(vertices, data=ptdata, crs=crs, copy_metadata=True)

def affine_matrix(mpa, mpb):
    """ Compute the affine transformation matrix that projects Multipoint mpa
    to Multipoint mpb using a least squares fit. """
    if len(mpa) != len(mpb):
        raise GeometryError("Input geometries must have identical length")
    vecp = np.asarray(mpb.get_vertices()).ravel()
    A = np.empty([2*len(mpa), 6], dtype=np.float64)
    for i, (x, y) in enumerate(mpa.get_vertices()):
        A[2*i:2*i+2,:] = np.kron(np.eye(2), [x, y, 1])
    M, res, rank, singvals = np.linalg.lstsq(A, vecp)
    return np.vstack([np.reshape(M, [2, 3]), np.atleast_2d([0, 0, 1])])

def sign(a):
    if a == 0.0:
        return 1
    else:
        return a/abs(a)



"""
2D raster functions
"""

import numpy as np

def witch_of_agnesi(nx=100, ny=100, a=4.0):
    """ Return a raster field defined by the equation
                    8 a^3
            Z = --------------
                 d^2 + 4 a*2
    where d is the distance from the center.
    """
    xc = int(np.floor(nx / 2.0))
    yc = int(np.floor(ny / 2.0))
    X, Y = np.meshgrid(range(nx), range(ny))
    D = np.sqrt( (X-xc)**2 + (Y-yc)**2 )

    return (8.0 * a**3) / (D**2 + 4 * a**2)


def peaks(n=49):
    """ 2d peaks function of MATLAB logo fame. """
    X, Y = np.meshgrid(np.linspace(-3, 3, n), np.linspace(-3, 3, n))
    return 3.0 * (1-X)**2 * np.exp(-X**2 - (Y+1)**2) \
            - 10.0 * (X/5.0 - X**3 - Y**5) * np.exp(-X**2 - Y**2) \
            - 1.0/3.0 * np.exp(-(X+1)**2 - Y**2)


def pad(A, width=1, edges="all", value=0.0):
    """ Apply padding to a 2D array.
            *A*         :   array to pad
            *width*     :   thickness of padding
            *edges*     :   "all", "left", "right", "top", "bottom"
            *value"     :   0.0, value to pad with
    """
    ny = A.shape[0]
    nx = A.shape[1]

    if edges == "all":
        edges = ["left", "right", "bottom", "top"]

    if "top" == edges or "top" in edges:
        ny += width
        y0 = width
    else:
        y0 = 0
    if "bottom" == edges or "bottom" in edges:
        ny += width
        yf = -width
    else:
        yf = None
    if "left" == edges or "left" in edges:
        nx += width
        x0 = width
    else:
        x0 = 0
    if "right" == edges or "right" in edges:
        nx += width
        xf = -width
    else:
        xf = None

    B = value * np.ones([ny, nx])
    B[y0:yf, x0:xf] = A

    return B


def slope(D, res=(30.0, 30.0)):
    """ Return the scalar slope at each pixel. Use the neighbourhood
    method.
    http://webhelp.esri.com/arcgisdesktop/9.2/index.cfm?TopicName=How%20Slope%20works
    """
    dx = res[0]
    dy = res[1]
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * dx)
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * dy)

    return pad(np.sqrt(Ddx*Ddx + Ddy*Ddy), value=np.nan)


def aspect(D, res=(30.0, 30.0)):
    """ Return the slope aspect for each pixel.
    http://webhelp.esri.com/arcgisdesktop/9.2/index.cfm?TopicName=How%20Aspect%20works
    """
    dx = res[0]
    dy = res[1]
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * dx)
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * dy)

    return pad(np.arctan2(Ddy, -Ddx), value=np.nan)


def grad(D, res=(30.0, 30.0)):
    """ Computes the gradient of potential D. Return a tuple (dx, dy).
    """
    dx = res[0]
    dy = res[1]
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * dx)
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * dy)
    return (pad(Ddx, width=1, edges='all', value=np.nan),
            pad(Ddy, width=1, edges='all', value=np.nan))


def div(U, V, res=(30.0, 30.0)):
    """ Calculate the divergence of a vector field. """
    dUdx = (U[:,2:] - U[:,:-2]) / (2.0*res[0])
    dVdy = (V[2:,:] - V[:-2,:]) / (2.0*res[1])
    divergence = pad(dUdx, width=1, edges=('left', 'right'), value=np.nan) \
               + pad(dVdy, width=1, edges=('top', 'bottom'), value=np.nan)
    return divergence


def normed_vector_field(D):
    """ Computes a U,V vector field of potential D. Scalar components of
    U,V are normalized to max(|U, V|).
    """
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / 8.0
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / 8.0
    M = np.sqrt(Ddx**2 + Ddy**2)
    U = Ddx / M[np.isnan(M)==False].max()
    V = Ddy / M[np.isnan(M)==False].max()
    return pad(U, value=np.nan), pad(V, value=np.nan)


def hillshade(D, res=(30.0, 30.0), bearing=330.0, azimuth=60.0):
    """ Return a hillshade raster for field D. Testing version. """
    dx, dy = grad(D, res=res)
    u = np.array((res[0] * np.ones_like(dx), np.zeros_like(dx), dx))
    v = np.array((np.zeros_like(dy), res[1] * np.ones_like(dy), dy))
    w = np.cross(u, v, axisa=0, axisb=0)
    wunit = w / np.atleast_3d(np.sqrt(np.sum(w**2, axis=-1)))
    pi = np.pi
    s = np.array((np.cos(bearing*pi/180.0),
                  np.sin(bearing*pi/180.0),
                  np.sin(azimuth*pi/180.0)))
    smat = s*np.ones([wunit.shape[0], wunit.shape[1], 3])
    dprod = (wunit*smat).sum(axis=-1)
    return dprod.T

def viewshed(D, i, j, r=-1):
    """ Return the viewshed on *D* at (i,j).
    """
    raise NotImplementedError



""" Convenience reader functions """

import os
import sys
from numbers import Number
import shapefile
from . import geometry
from . import geojson
from . import xyfile
from .metadata import Metadata
from ..crs import LonLatWGS84, HASOSR, crs_from_wkt
from .. import errors

def from_shape(obj, properties=None):
    """ Read a __geo_interface__ dictionary and return an appropriate karta
    object """
    return _from_shape(obj.__geo_interface__, None)

def _from_shape(d, properties):
    if d["type"] == "Feature":
        p = d["properties"]
        return _from_shape(d["geometry"], p)
    else:
        c = d["coordinates"]
        if d["type"] == "Point":
            return geometry.Point(c, properties=properties)
        elif d["type"] == "MultiPoint":
            return geometry.Multipoint(c, properties=properties)
        elif d["type"] == "LineString":
            return geometry.Line(c, properties=properties)
        elif d["type"] == "Polygon":
            subs = [geometry.Polygon(c[i]) for i in range(1, len(c))]
            return geometry.Polygon(c[0], properties=properties, subs=subs)
        elif d["type"] == "MultiLineString":
            return [geometry.Line(c[i], properties=properties)
                    for i in range(len(c))]
        elif d["type"] == "MultiPolygon":
            polys = []
            for c_ in c:
                subs = [geometry.Polygon(c_[i]) for i in range(1, len(c_))]
                polys.append(geometry.Polygon(c_[0], properties=properties,
                                              subs=subs))
            return polys
        else:
            raise NotImplementedError("Geometry type {0} not "
                                      "implemented".format(d["type"]))

def read_geojson(f):
    """ Read a GeoJSON file object and return a list of geometries """

    def convert_crs(crsdict):
        if crsdict.get("type", None) not in ("name", "link"):
            crs = LonLatWGS84
        elif crsdict["type"] == "name":
            crs = geojson.GeoJSONNamedCRS(crsdict["properties"]["name"])
        elif crsdict["type"] == "link":
            crs = geojson.GeoJSONLinkedCRS(crsdict["properties"]["href"],
                                           crsdict["properties"]["type"])
        return crs

    def convert(geom, **kw):
        if isinstance(geom, geojson.Feature):
            res = [convert_feature(geom)]
        elif isinstance(geom, geojson.FeatureCollection):
            res = [convert_feature(f, **kw) for f in geom.features]
        elif isinstance(geom, geojson.GeometryCollection):
            res = [convert(item, **kw) for item in geom.geometries]
        else:
            res = convert_geometry(geom, **kw)
        return res

    def convert_geometry(geom, **kw):
        crs = convert_crs(geom.crs)
        if isinstance(geom, geojson.Point):
            return geometry.Point(geom.coordinates, crs=crs, **kw)
        elif isinstance(geom, geojson.MultiPoint):
            return geometry.Multipoint(geom.coordinates, crs=crs, **kw)
        elif isinstance(geom, geojson.LineString):
            return geometry.Line(geom.coordinates, crs=crs, **kw)
        elif isinstance(geom, geojson.Polygon):
            return geometry.Polygon(geom.coordinates[0],
                                    subs=geom.coordinates[1:],
                                    crs=crs, **kw)
        else:
            raise TypeError("{0} is a not a JSON geometry".format(geom))

    def convert_feature(feat, **kw):
        data = feat.properties["vector"]
        if len(data) == 0:
            data = None
        else:
            for key, val in data.items():
                if any(isinstance(a, Number) or hasattr(a, "dtype") for a in val):
                    for i in range(len(val)):
                        if val[i] is None:
                            val[i] = float('nan')
        prop = feat.properties["scalar"]
        kw["data"] = data
        kw["properties"] = prop
        return convert(feat.geometry, **kw)

    R = geojson.GeoJSONReader(f)
    geom = R.parse()
    return convert(geom)

def _geojson_properties2karta(properties, n):
    """ Takes a dictionary (derived from a GeoJSON properties object) and
    divides it into singleton properties and *n*-degree data. """
    props = {}
    data = {}
    for (key, value) in properties.items():
        if isinstance(value, list) or isinstance(value, tuple):
            if len(value) == n:
                data[key] = value
            else:
                raise ValueError("properties must be singleton or per-vertex")
        else:
            props[key] = value
    return props, data

def read_xyfile(f, delimiter='', header_rows=0, astype=geometry.Multipoint, coordrank=2):
    """ Read an ASCII delimited table and return a geometry object given by *astype*.
    """
    dat = xyfile.load_xy(f, delimiter=delimiter, header_rows=header_rows)
    ncols = dat.shape[1]
    if ncols >= coordrank:
        coords = dat[:,:coordrank]
        if ncols > coordrank:
            data = dat[:,coordrank:]
        else:
            data = None
        return astype(coords, data=data)
    else:
        raise IOError('data table has insufficient number of columns')

### Shapefile functions ###

def get_filenames(stem, check=False):
    """ Given a filename basename, return the associated shapefile paths. If
    `check` is True, ensure that the files exist."""
    shp = stem + '.shp'
    shx = stem + '.shx'
    dbf = stem + '.dbf'
    if check:
        for fnm in (shp, shx, dbf):
            if not os.path.isfile(fnm):
                raise Exception('missing {0}'.format(fnm))
    return {'shp':shp, 'shx':shx, 'dbf':dbf}

def open_file_dict(fdict):
    """ Open each file in a dictionary of filenames and return a matching
    dictionary of the file objects. """
    files = {}
    for ext in fdict.keys():
        files[ext] = open(fdict[ext], 'rb')
    return files

dBase_type_dict = {"I": int,
                   "O": float,
                   "C": str,
                   "@": lambda a: a,    # Temporary
                   "D": lambda a: a,
                   "L": bool}

def recordsasdata(reader):
    """ Interpret shapefile records as a Metadata object """
    d = {}
    idfunc = lambda a: a
    try:
        records = [rec for rec in reader.records()]
        for (i,k) in enumerate(reader.fields[1:]):
            f = dBase_type_dict.get(k[1], idfunc)
            d[k[0]] = [f(rec[i]) for rec in records]
    except ValueError:
        # bug in QGIS/pyshp, fixed in dev version of pyshp but not on PyPI yet
        # see pyshp commit d72723c9e38db8e859b79d95a65c00af1c2ba8ba
        sys.stderr.write("Warning: failed to read records (see "
                         "https://github.com/GeospatialPython/pyshp/pull/16)\n")
        for (i,k) in enumerate(reader.fields[1:]):
            f = dBase_type_dict.get(k[1], idfunc)
            d[k[0]] = [None for i in range(reader.numRecords)]
    return Metadata(d)

def recordsasproperties(reader):
    """ Interpret shapefile records as a list of properties dictionaries """
    proplist = []
    keys = reader.fields[1:]
    idfunc = lambda a: a
    try:
        for (i,rec) in enumerate(reader.records()):
            properties = {}
            for (k,v) in zip(keys, rec):
                f = dBase_type_dict.get(k[1], idfunc)
                properties[k[0]] = f(v)
            proplist.append(properties)
    except ValueError:
        # bug in QGIS/pyshp, fixed in dev version of pyshp but not on PyPI yet
        # see pyshp commit d72723c9e38db8e859b79d95a65c00af1c2ba8ba
        sys.stderr.write("Warning: failed to read records (see "
                         "https://github.com/GeospatialPython/pyshp/pull/16)\n")
        proplist = [None for i in range(reader.numRecords)]
    return proplist

def crs_from_prj(prjfnm):
    """ Read a *.prj file and return a matching CRS instance """
    if not HASOSR:
        raise errors.MissingDependencyError("Parsing prj files is performed by osgeo.osr, which is not installed.")

    with open(prjfnm, "r") as f:
        wkt_str = f.read()
    return crs_from_wkt(wkt_str)

def read_shapefile(name, crs=None):
    """ Read a shapefile given `name`, which may include or exclude the .shp
    extension. The CRS must be specified, otherwise it is assumed to be
    cartesian. """
    if os.path.splitext(name)[1] == ".shp":
        name = name[:-4]
    fnms = get_filenames(name, check=True)

    # Construct a CRS instance
    if crs is None:
        prjfnm = name + ".prj"
        qpjfnm = name + ".qpj"

        if os.path.exists(prjfnm):
            try:
                crs = crs_from_prj(prjfnm)
            except errors.CRSError as e:
                # attempt to recover with a warning
                sys.stderr.write(str(e))
                crs = LonLatWGS84

        elif os.path.exists(qpjfnm):
            raise NotImplementedError("Currently support for *.QPJ projection data is missing. Specify CRS explicitly with the 'crs' keyword")

        else:
            # Make a guess -- however it seems this could be harmful
            crs = LonLatWGS84

    # Read shapefile data
    try:
        files = open_file_dict(fnms)
        reader = shapefile.Reader(shp=files['shp'], shx=files['shx'],
                                  dbf=files['dbf'])

        if reader.shapeType in (1, 11, 21):       # Points
            verts = [shp.points[0] for shp in reader.shapes()]
            d = recordsasdata(reader)
            geoms = [geometry.Multipoint(verts, data=d, crs=crs)]

        elif reader.shapeType in (3, 13, 23):     # Lines
            plist = recordsasproperties(reader)
            geoms = []
            for (shp,prop) in zip(reader.shapes(), plist):
                parts = list(shp.parts)
                parts.append(len(shp.points))
                for (i0, i1) in zip(parts[:-1], parts[1:]):
                    points = shp.points[i0:i1]
                    geoms.append(geometry.Line(points, properties=prop, crs=crs))

        elif reader.shapeType in (5, 15, 25):     # Polygon
            plist = recordsasproperties(reader)
            geoms = []
            for (shp,prop) in zip(reader.shapes(), plist):
                parts = list(shp.parts)
                parts.append(len(shp.points))
                for (i0, i1) in zip(parts[:-1], parts[1:]):
                    # Shapefile polygons are closed explicitly, while karta
                    # polygons are closed implicitly
                    points = shp.points[i0:i1-1]
                    geoms.append(geometry.Polygon(points, properties=prop, crs=crs))

        else:
            raise NotImplementedError("Shapefile shape type {0} not "
                "implemented".format(reader.shapeType))

    finally:
        for f in files.values():
            f.close()

    return geoms




""" Implements a simple quadtree datastructure, with emphasis on performance. """

from ._cvectorgeo import iswithin, hashpt

class Node(object):
    def __init__(self, children, bbox, leaf):
        self.children = children
        self.bbox = bbox
        self.leaf = leaf

class QuadTree(object):
    """ Implements a convenience class that wraps a quadtree data structure and
    the methods needed to grow or query it.

    Initialize with bbox (xmin, xmax, ymin, ymax).

    Optional parameters
    -------------------
    maxchildren         <int> maximum number of points before a node is split
    maxdepth            <int> maximum tree depth
    """

    def __init__(self, bbox, maxchildren=20, maxdepth=999):

        self.maxchildren = maxchildren
        self.maxdepth = maxdepth
        self.node = Node([], bbox, True)
        self.size = 0

    def addpt(self, pt):
        """ Add point to the QuadTree. Returns the depth at which the point was
        added. """
        (self.node, d) = addpt(self.node, pt, 1, self.maxchildren, self.maxdepth)
        self.size += 1
        return d

    def querypt(self, pt, method="hash"):
        """ Test whether QuadTree contains a point. *method* may be "recursion"
        [default] or "hash". """
        if method == "recursion":
            return querypt_recursion(self.node, pt)
        elif method == "hash":
            return querypt_hash(self.node, pt)
        else:
            raise ValueError("Unrecognized method '{0}'".format(method))

    def getfrombbox(self, bbox):
        """ Extract all points from a boundary box. """
        return getfrombbox(self.node, bbox)


def addpt(node, pt, depth, maxchildren, maxdepth):
    """ Add *pt* to *node*. *depth* is the current level in the quadtree.
    Returns the updated node (which may or may not be the same as the starting
    node, depending on whether a split occurred, as well as the depth at which
    the point was inserted. """
    if not iswithin(node.bbox, pt):
        raise BBoxError("({x},{y}) is not within bounds {bbox}".format(x=pt[0],
                                                                       y=pt[1],
                                                                       bbox=node.bbox))

    if node.leaf:

        if len(node.children) == maxchildren and depth != maxdepth:
            node = split(node)
            (node, depth) = addpt(node, pt, depth, maxchildren, maxdepth)

        else:
            node.children.append(pt)

    else:

        for (i,child) in enumerate(node.children):
            if iswithin(child.bbox, pt):
                (child, depth) = addpt(child, pt, depth+1, maxchildren, maxdepth)
                node.children[i] = child
                break

    return node, depth

def mean(x):
    n = len(x)
    if n != 0:
        return sum(x) / n
    else:
        None

def overlaps(bbox0, bbox1):
    pts0 = ((x, y) for x in bbox0[:2] for y in bbox0[2:])
    if any(iswithin(bbox1, pt) for pt in pts0):
        return True
    pts1 = ((x, y) for x in bbox1[:2] for y in bbox1[2:])
    if any(iswithin(bbox0, pt) for pt in pts1):
        return True
    return False

def split(node):
    """ Return a new node with the bbox of *node* and *node*'s children split
    between four child nodes. """
    (x0, x1, y0, y1) = node.bbox
    xm = 0.5 * (x0+x1)
    ym = 0.5 * (y0+y1)

    branch = Node([], node.bbox, False)
    bboxes = ((x0, xm, y0, ym), (xm, x1, y0, ym),
              (x0, xm, ym, y1), (xm, x1, ym, y1))

    for bbox in bboxes:
        pts = [pt for pt in node.children if iswithin(bbox, pt)]
        child = Node(pts, bbox, True)
        branch.children.append(child)
    return branch

def getfrombbox(parent, bbox):
    """ Return all points in *parent* node that are in *bbox*. """
    if parent.leaf:
        return [pt for pt in parent.children if iswithin(bbox, pt)]
    else:
        pts = []
        for child in parent.children:
            if overlaps(bbox, child.bbox):
                pts.extend(getfrombbox(child, bbox))
        return pts

def querypt_recursion(parent, pt):
    """ Test whether a point exists at *pt* using recursion. """
    if parent.leaf:
        for childpt in parent.children:
            if pt == childpt:
                return True
    else:
        for child in parent.children:
            if iswithin(child.bbox, pt):
                return querypt_recursion(child, pt)
    return False

def querypt_hash(parent, pt):
    (a, b, c, d) = parent.bbox
    h = hashpt(a, b, c, d, pt[0], pt[1])
    for quad in h:
        if parent.leaf:
            return pt in parent.children
        else:
            parent = parent.children[quad]

class BBoxError(Exception):
    def __init__(self, message):
        self.message = message
    def __repr__(self):
        return self.message



""" Defines basic geodetic operations on a planes and spheres. """

#
# Functions are defined with an underscored version that works on scalars and a
# public version that dispatches between scalars and vectors. This approach is
# used rather than vectorization to make the C translation more straighforward.
#

from __future__ import division
import numpy as np
from math import sqrt, sin, cos, tan, asin, acos, atan, atan2, atanh, pi
import warnings

# ---------------------------------
# Functions for planar geometry
# ---------------------------------

def _plane_distance(x1, y1, x2, y2):
    return sqrt((x2 - x1)**2 + (y2 - y1)**2)

def plane_distance(xs1, ys1, xs2, ys2):
    if hasattr(xs1, "__len__"):
        return np.array([_plane_distance(x1, y1, x2, y2)
                        for x1, y1, x2, y2 in zip(xs1, ys1, xs2, ys2)])
    else:
        return _plane_distance(xs1, ys1, xs2, ys2)

def _plane_azimuth(x1, y1, x2, y2):
    """ Return cartesian azimuth between points (scalar func) """
    dx = x2 - x1
    dy = y2 - y1
    if dx > 0:
        if dy > 0:
            return np.arctan(dx/dy)
        elif dy < 0:
            return np.pi - np.arctan(-dx/dy)
        else:
            return 0.5*np.pi
    elif dx < 0:
        if dy > 0:
            return 2*np.pi - np.arctan(-dx/dy)
        elif dy < 0:
            return np.pi + np.arctan(dx/dy)
        else:
            return 1.5*np.pi
    else:
        if dy > 0:
            return 0.0
        else:
            return np.pi

def plane_azimuth(xs1, ys1, xs2, ys2):
    """ Return cartesian azimuth between points """
    if hasattr(xs1, "__len__"):
        return np.array([_plane_azimuth(x1, y1, x2, y2)
                        for x1, y1, x2, y2 in zip(xs1, ys1, xs2, ys2)])
    else:
        return _plane_azimuth(xs1, ys1, xs2, ys2)


# ---------------------------------
# Functions for spherical geodesy
# ---------------------------------

def _sphere_distance(lon1, lat1, lon2, lat2, radius):
    dx = abs(lon1-lon2)
    dy = abs(lat1-lat2)

    if dx > 0.01 or dy > 0.01:
        # use spherical law of cosines
        d_ = acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(dx))

    else:
        # use haversine
        d_ = (2 * asin(sqrt(sin(dy / 2.)**2 +
                        cos(lat1) * cos(lat2) * sin(dx / 2.)**2)))
    return radius * d_

def sphere_distance(lons1, lats1, lons2, lats2, radius):
    if hasattr(lons1, "__len__"):
        return np.array([_sphere_distance(ln1, lt1, ln2, lt2, radius)
                        for ln1, lt1, ln2, lt2
                        in zip(lons1, lats1, lons2, lats2)])
    else:
        return _sphere_distance(lons1, lats1, lons2, lats2, radius)


def _sphere_azimuth(lon1, lat1, lon2, lat2):
    dlon = lon2 - lon1
    if (cos(lat1) * tan(lat2) - sin(lat1) * cos(dlon)) == 0:
        az = 0.5*pi
    else:
        az = atan(sin(dlon) / (cos(lat1) * tan(lat2) - sin(lat1) * cos(dlon)))
    
    if unroll_angle(dlon) <= pi:
        if lat2 < lat1:
            az = az + pi
    else:
        if lat1 < lat2:
            az = az + 2*pi
        else:
            az = az + pi
    return unroll_angle(az)

def sphere_azimuth(lons1, lats1, lons2, lats2):
    if hasattr(lons1, "__len__"):
        return np.array([_sphere_azimuth(lon1, lat1, lon2, lat2)
                        for lon1, lat1, lon2, lat2
                        in zip(lons1, lats1, lons2, lats2)])
    else:
        return _sphere_azimuth(lons1, lats1, lons2, lats2)

def _spherical_area(r, alpha1, alpha2):
    return r**2*dalpha


def spherical_area(r, x1, y1, x2, y2):
    _, x1, y1, x2, y2 = _canonical_configuration(x1, y1, x2, y2)
    phi1 = y1*pi/180.0
    phi2 = y2*pi/180.0
    lambda12 = (x2-x1)*pi/180.0
    alpha1, alpha2, _ = solve_vicenty(r, 0, lambda12, phi1, phi2)
    return r**2 * (alpha2-alpha1)

# ---------------------------------
# Functions for ellipsoidal geodesy
# ---------------------------------

def solve_astroid(a, f, lambda12, phi1, phi2):
    """ Used to provide an initial guess to the inverse problem in the case of
    nearly antipodal points.

    Parameters
    ----------
    a:          equatorial radius
    f:          flattening
    lambda12:   difference in longitudes (radians)
    phi1:       first latitude (radians)
    phi2:       second latitude (radians)

    Returns
    -------
    alpha1:     estimated forward azimuth at first point

    see Karney (2013) J. Geod. for details
    """
    beta1 = atan((1-f) * tan(phi1))
    beta2 = atan((1-f) * tan(phi2))
    delta = f*a*pi*cos(beta1)**2
    x = (lambda12-pi) * (a*cos(beta1)) / delta
    y = (beta2 + beta1) * a / delta
    mu = fzero_brent(1e-3, pi*a, lambda mu: mu**4 + 2*mu**3 + (1-x**2-y**2)*mu**2 - 2*y**2*mu - y**2, 1e-12)
    alpha1 = atan2(-x / (1+mu), y/mu)
    return alpha1

def solve_vicenty(a, f, lambda12, phi1, phi2):
    """ Used to provide an initial guess to the inverse problem by solving the
    corresponding problem on a sphere.

    Parameters
    ----------
    a:          equatorial radius
    f:          flattening
    lambda12:   difference in longitudes (radians)
    phi1:       first latitude (radians)
    phi2:       second latitude (radians)

    Returns
    -------
    alpha1:     forward azimuth at first point
    alpha2:     forward azimuth at second point
    s12:        distance between points

    see Karney (2013) J. Geod. for details
    """
    eccn2 = f*(2-f)
    beta1 = atan((1-f) * tan(phi1))
    beta2 = atan((1-f) * tan(phi2))
    w = sqrt(1 - eccn2 * (0.5 * (cos(beta1) + cos(beta2)))**2)
    omega12 = lambda12 / w

    z1_r = cos(beta1)*sin(beta2) - sin(beta1)*cos(beta2)*cos(omega12)
    z1_i = cos(beta2)*sin(omega12)
    z1 = sqrt(z1_r**2 + z1_i**2)
    sigma12 = atan2(z1, sin(beta1)*sin(beta2) + cos(beta1)*cos(beta2)*cos(omega12))
    z2_r = -sin(beta1)*cos(beta2) + cos(beta1)*sin(beta2)*cos(omega12)
    z2_i = cos(beta1)*sin(omega12)

    alpha1 = atan2(z1_i, z1_r)
    alpha2 = atan2(z2_i, z2_r)
    s12 = a*w*sigma12
    return alpha1, alpha2, s12

def _solve_NEA(alpha0, alpha1, beta1):
    """ See Karney (2013) """
    sigma1 = atan2(sin(beta1), cos(alpha1)*cos(beta1))
    omega1 = atan2(sin(alpha0)*sin(sigma1), cos(sigma1))
    return sigma1, omega1

def _solve_NEB(alpha0, alpha1, beta1, beta2):
    """ See Karney (2013) """
    try:
        alpha2 = acos(sqrt(cos(alpha1)**2*cos(beta1)**2 + (cos(beta2)**2 - cos(beta1)**2)) / cos(beta2))
    except ValueError:
        alpha2 = asin(sin(alpha0) / cos(beta2))     # Less accurate?
    sigma2 = atan2(sin(beta2), cos(alpha2)*cos(beta2))
    omega2 = atan2(sin(alpha0)*sin(sigma2), cos(sigma2))
    return alpha2, sigma2, omega2

def _normalize_longitude(x):
    """ Return longitude in the range [-180, 180). """
    return (x+180) % 360 - 180

def _canonical_configuration(x1, y1, x2, y2):
    """ Put coordinates into a configuration where (Karney, eqn 44)
        y1 <= 0
        y1 <= y2 <= -y1
        0 <= x2-x1 <= 180
    """
    transformation = dict(yflip=False, xflip=False, ysignswap=False)

    if abs(y1) < abs(y2):
        y1, y2 = y2, y1
        transformation["yflip"] = True

    if y1 > 0:
        y1, y2 = -y1, -y2
        transformation["ysignswap"] = True

    x2 = _normalize_longitude(x2-x1)
    x1 = 0.0

    if (x2 < 0) or (x2 > 180):
        x2 = -x2
        transformation["xflip"] = True

    return transformation, x1, y1, x2, y2

def ellipsoidal_forward(a, b, x, y, azimuth, distance):
    """ Compute the destination reached starting from a point and travelling
    in a specified direction.

    Parameters
    ----------
    a:          equatorial radius
    b:          polar radius
    x:          longitude at start
    y:          latitude at start
    azimuth:    direction travelled from point
    distnce:    distance travelled

    Returns
    -------
    x2:         longitude at destination
    y2:         latitude at destination
    back_az:    back azimuth from destination

    Algorithm due to Karney, C.F.F. "Algorithms for geodesics", J. Geod (2013)
    """
    f = (a-b) / a

    phi1 = pi*y/180.0
    alpha1 = pi*azimuth/180.0

    # Solve triangle NEA from Karney Fig. 1
    beta1 = atan((1-f)*tan(phi1))
    _i = sqrt(cos(alpha1)**2 + (sin(alpha1)*sin(beta1))**2)
    alpha0 = atan2(sin(alpha1)*cos(beta1), _i)
    # sigma1 = atan2(sin(beta1), cos(alpha1)*cos(beta1))
    # omega1 = atan2(sin(alpha0)*sin(sigma1), cos(sigma1))
    sigma1, omega1 = _solve_NEA(alpha0, alpha1, beta1)

    # Determine sigma2
    eccn2 = (f*(2-f))
    second_eccn2 = eccn2 / (1-eccn2)
    k2 = second_eccn2*cos(alpha0)**2

    _rad = sqrt(1+k2)
    eps = (_rad - 1) / (_rad + 1)
    A1 = 1.0/(1-eps) * (1 + eps**2/4 + eps**4/64 + eps**6/256)
    C1 = [-1.0/2*eps + 3.0/16*eps**3 - 1.0/32*eps**5,
          -1.0/16*eps**2 + 1.0/32*eps**4 - 9.0/2048*eps**6,
          -1.0/48*eps**3 + 3.0/256*eps**5,
          -5.0/512*eps**4 + 3.0/512*eps**6,
          -7.0/1280*eps**5,
          -7.0/2048*eps**6]

    I1 = A1 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C1)))
    s1 = I1 * b
    s2 = s1 + distance
    tau2 = s2 / (b*A1)

    C1p = [eps/2 - 9.0/32*eps**3 + 205.0/1536*eps**5,
           5.0/16*eps**2 - 37.0/96*eps**4 + 1335.0/4096*eps**6,
           29.0/96*eps**3 - 75.0/128*eps**5,
           539.0/1536*eps**4 - 2391.0/2560*eps**6,
           3467.0/7680*eps**5,
           38081.0/61440*eps**6]

    sigma2 = tau2 + sum(c*sin(2*(i+1)*tau2) for i,c in enumerate(C1p))

    # Solve triangle NEB in Karney Fig. 1
    alpha2 = atan2(sin(alpha0), cos(alpha0)*cos(sigma2))
    _j = sqrt((cos(alpha0)*cos(sigma2))**2 + sin(alpha0)**2)
    beta2 = atan2(cos(alpha0)*sin(sigma2), _j)
    omega2 = atan2(sin(alpha0)*sin(sigma2), cos(sigma2))

    # Determine lambda12
    n = f / (2.0-f)
    n2 = n*n
    A3 = 1.0 - (1.0/2-n/2)*eps - (1.0/4 + n/8 - 3.0*n2/8)*eps**2 \
        - (1.0/16 + 3.0*n/16 + n2/16)*eps**3 - (3.0/64 + n/32)*eps**4 \
        - 3.0/128*eps**5

    C3 = [(1.0/4 - n/4)*eps + (1.0/8 - n2/8)*eps**2 + (3.0/64 + 3.0*n/64 - n2/64)*eps**3 \
            + (5.0/128 + n/64)*eps**4 + 3.0/128*eps**5,
          (1.0/16 - 3.0*n/32 + n2/32)*eps**2 + (3.0/64 - n/32 - 3*n2/64)*eps**3 \
            + (3.0/128 + n/128)*eps**4 + 5.0/256*eps**5,
          (5.0/192 - 3.0*n/64 + 5.0*n2/192)*eps**3 + (3.0/128 - 5.0*n/192)*eps**4 \
            + 7.0/512*eps**5,
          (7.0/512 - 7.0*n/256)*eps**4 + 7.0*eps**5/512,
          21.0*eps**5/2560]

    I3s1 = A3 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C3)))
    I3s2 = A3 * (sigma2 + sum(c*sin(2*(i+1)*sigma2) for i,c in enumerate(C3)))

    lambda1 = omega1 - f*sin(alpha0)*I3s1
    lambda2 = omega2 - f*sin(alpha0)*I3s2
    lambda12 = lambda2 - lambda1

    phi2 = atan(tan(beta2) / (1-f))
    x2 = x + lambda12*180.0/pi
    if x2 >= 180.0:
        x2 -= 360.0
    y2 = phi2*180.0/pi
    backaz = (alpha2+pi)*180/pi
    x2 = (x2+180) % 360 - 180
    backaz = (backaz+180) % 360 - 180
    return x2, y2, backaz

def _ellipsoidal_inverse_equatorial(a, x1, x2):
    diff = (x2-x1 + 180) % 360 - 180
    if diff < 0:
        az = -90.0
        baz = 90.0
    else:
        az = 90.0
        baz = -90.0
    s12 = 2 * pi * a * abs(x1-x2)/360.0
    return az, baz, s12

def ellipsoidal_inverse(a, b, x1, y1, x2, y2, tol=None):
    """ Compute the shortest path (geodesic) between two points.

    Parameters
    ----------
    a:          equatorial radius
    b:          polar radius
    x1:         first longitude
    y1:         first latitude
    x2:         second longitude
    y2:         second latitude

    Returns
    -------
    az:         forward azimuth from first point
    back_az:    back azimuth from second point
    distance:   distance between points

    Algorithm due to Karney, C.F.F. "Algorithms for geodesics", J. Geod (2013)
    """
    niter = 0
    maxiter = 100
    if tol is None:
        tol = 1e-12

    if y1 == y2 == 0:
        # Equatorial case
        return _ellipsoidal_inverse_equatorial(a, x1, x2)

    # Canonical configuration
    tr, x1, y1, x2, y2 = _canonical_configuration(x1, y1, x2, y2)

    phi1 = y1*pi/180.0
    phi2 = y2*pi/180.0
    lambda12 = (x2-x1)*pi/180.0
    f = (a-b) / a

    beta1 = atan((1-f)*tan(phi1))
    beta2 = atan((1-f)*tan(phi2))

    eccn2 = f*(2-f)
    second_eccn2 = eccn2 / (1-eccn2)

    if x1 == x2:
        # Meridional case 1
        alpha0 = alpha1 = alpha2 = omega1 = omega2 = 0.0

        _i = sqrt(cos(alpha1)**2 + (sin(alpha1)*sin(beta1))**2)
        alpha0 = atan2(sin(alpha1)*cos(beta1), _i)
        sigma1, _ = _solve_NEA(alpha0, alpha1, beta1)
        _, sigma2, _ = _solve_NEB(alpha0, alpha1, beta1, beta2)

        k2 = second_eccn2
        _rad = sqrt(1+k2)
        eps = (_rad - 1) / (_rad + 1)

    elif abs(lambda12 % (2*pi) - pi) < 1e-12:
        # Meridional case 2
        if y1 + y2 > 0:
            alpha0 = alpha1 = 0.0
            alpha2 = omega1 = omega2 = pi
        else:
            alpha0 = alpha1 = omega1 = omega2 = pi
            alpha2 = 0.0

        sigma1, _ = _solve_NEA(alpha0, alpha1, beta1)
        _, sigma2, _ = _solve_NEB(alpha0, alpha1, beta1, beta2)

        k2 = second_eccn2
        _rad = sqrt(1+k2)
        eps = (_rad - 1) / (_rad + 1)

    else:
        # Newton iteration

        # Guess the azimuth
        if (abs(lambda12-pi) > 0.0087) and (abs(phi1+phi2) > 0.0087):
            # not nearly antipodal
            alpha1, _, _ = solve_vicenty(a, f, lambda12, phi1, phi2)
        else:
            alpha1 = solve_astroid(a, f, lambda12, phi1, phi2)

        dlambda12 = tol + 1

        while (abs(dlambda12) > tol) and (niter != maxiter):

            # Solve triangles
            _i = sqrt(cos(alpha1)**2 + (sin(alpha1)*sin(beta1))**2)
            alpha0 = atan2(sin(alpha1)*cos(beta1), _i)
            sigma1, omega1 = _solve_NEA(alpha0, alpha1, beta1)
            alpha2, sigma2, omega2 = _solve_NEB(alpha0, alpha1, beta1, beta2)

            # Determine lambda12
            k2 = second_eccn2 * cos(alpha0)**2
            _rad = sqrt(1+k2)
            eps = (_rad - 1) / (_rad + 1)

            n = f/(2-f)
            n2 = n*n
            A3 = 1.0 - (1.0/2 - 1.0/2*n)*eps - (1.0/4 + 1.0/8*n - 3.0/8*n2)*eps**2 \
                - (1.0/16 + 3.0/16*n + 1.0/16*n2)*eps**3 - (3.0/64 + 1.0/32*n)*eps**4 \
                - 3.0/128*eps**5

            C3 = [(1.0/4 - n/4)*eps + (1.0/8 - n2/8)*eps**2 + (3.0/64 + 3.0*n/64 - n2/64)*eps**3 \
                    + (5.0/128 + n/64)*eps**4 + 3.0/128*eps**5,
                  (1.0/16 - 3.0*n/32 + n2/32)*eps**2 + (3.0/64 - n/32 - 3*n2/64)*eps**3 \
                    + (3.0/128 + n/128)*eps**4 + 5.0/256*eps**5,
                  (5.0/192 - 3.0*n/64 + 5.0*n2/192)*eps**3 + (3.0/128 - 5.0*n/192)*eps**4 \
                    + 7.0/512*eps**5,
                  (7.0/512 - 7.0*n/256)*eps**4 + 7.0*eps**5/512,
                  21.0*eps**5/2560]

            I3s1 = A3 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C3)))
            I3s2 = A3 * (sigma2 + sum(c*sin(2*(i+1)*sigma2) for i,c in enumerate(C3)))

            lambda1 = omega1 - f*sin(alpha0)*I3s1
            lambda2 = omega2 - f*sin(alpha0)*I3s2
            lambda12_next = lambda2 - lambda1
            dlambda12 = lambda12_next - lambda12

            if abs(dlambda12) > tol:
                # Refine alpha1
                A1 = 1.0/(1-eps) * (1 + eps**2/4 + eps**4/64 + eps**6/256)
                C1 = [-1.0/2*eps + 3.0/16*eps**3 - 1.0/32*eps**5,
                      -1.0/16*eps**2 + 1.0/32*eps**4 - 9.0/2048*eps**6,
                      -1.0/48*eps**3 + 3.0/256*eps**5,
                      -5.0/512*eps**4 + 3.0/512*eps**6,
                      -7.0/1280*eps**5,
                      -7.0/2048*eps**6]

                I1s1 = A1 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C1)))
                I1s2 = A1 * (sigma2 + sum(c*sin(2*(i+1)*sigma2) for i,c in enumerate(C1)))

                A2 = (1-eps) * (1 + 1.0/4*eps**2 + 9.0/64*eps**4 + 25.0/256*eps**6)
                C2 = [1.0/2*eps + 1.0/16*eps**3 + 1.0/32*eps**5,
                      3.0/16*eps**2 + 1.0/32*eps**4 + 35.0/2048*eps**6,
                      5.0/48*eps**3 + 5.0/256*eps**5,
                      35.0/512*eps**4 + 7.0/512*eps**6,
                      63.0/1280*eps**5,
                      77.0/2048*eps**6]

                I2s1 = A2 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C2)))
                I2s2 = A2 * (sigma2 + sum(c*sin(2*(i+1)*sigma2) for i,c in enumerate(C2)))

                Js1 = I1s1 - I2s1
                Js2 = I1s2 - I2s2

                m12 = b * (sqrt(1 + k2*sin(sigma2)**2) * cos(sigma1)*sin(sigma2) \
                         - sqrt(1 + k2*sin(sigma1)**2) * sin(sigma1)*cos(sigma2) \
                         - cos(sigma1) * cos(sigma2) * (Js2-Js1))
                dlambda12_dalpha1 = m12/(a * cos(alpha2)*cos(beta2))
                dalpha1 = -dlambda12 / dlambda12_dalpha1
                alpha1 = (alpha1 + dalpha1) % (2*pi)

            niter += 1

    if niter == maxiter:
        warnings.warn(
            "Convergence failure (%f, %f) -> (%f, %f)" % (x1, y1, x2, y2), 
            RuntimeWarning)

    k2 = second_eccn2 * cos(alpha0)**2
    _rad = sqrt(1+k2)
    eps = (_rad - 1) / (_rad + 1)

    # Determine s12
    A1 = 1.0/(1-eps) * (1 + eps**2/4 + eps**4/64 + eps**6/256)
    C1 = [-1.0/2*eps + 3.0/16*eps**3 - 1.0/32*eps**5,
          -1.0/16*eps**2 + 1.0/32*eps**4 - 9.0/2048*eps**6,
          -1.0/48*eps**3 + 3.0/256*eps**5,
          -5.0/512*eps**4 + 3.0/512*eps**6,
          -7.0/1280*eps**5,
          -7.0/2048*eps**6]

    I1s1 = A1 * (sigma1 + sum(c*sin(2*(i+1)*sigma1) for i,c in enumerate(C1)))
    I1s2 = A1 * (sigma2 + sum(c*sin(2*(i+1)*sigma2) for i,c in enumerate(C1)))

    s1 = I1s1*b
    s2 = I1s2*b
    s12 = s2-s1

    if tr["xflip"]:
        alpha1 = -alpha1
        alpha2 = -alpha2

    if tr["yflip"]:
        alpha1, alpha2 = pi-alpha2, pi-alpha1

    if tr["ysignswap"]:
        alpha1 = pi - alpha1
        alpha2 = pi - alpha2

    az = (alpha1*180/pi + 180) % 360 - 180
    backaz = ((alpha2+pi)*180/pi + 180) % 360 - 180
    return az, backaz, s12

def _ellipsoidal_area(a, b, lambda12, phi1, phi2, alpha1, alpha2):
    """ Area of a single quatrilateral defined by two meridians, the equator,
    and another geodesic.
    """
    f = (a-b)/a
    e2 = f*(2-f)
    ep2 = e2/(1-e2)
    e = sqrt(e2)

    # Authalic radius
    c = sqrt(a**2/2 + b**2/2*atanh(e)/e)

    beta1 = atan((1-f)*tan(phi1))
    beta2 = atan((1-f)*tan(phi2))

    _i = sqrt(cos(alpha1)**2 + (sin(alpha1)*sin(beta1))**2)
    alpha0 = atan2(sin(alpha1)*cos(beta1), _i)

    sigma1, omega1 = _solve_NEA(alpha0, alpha1, beta1)
    _, sigma2, omega2 = _solve_NEB(alpha0, alpha1, beta1, beta2)
    omega12 = omega2 - omega1

    # Bessel identity for alpha2 - alpha1
    alpha12 = 2*atan(sin(0.5*beta1+0.5*beta2)/cos(0.5*beta2-0.5*beta1) \
            * tan(0.5*omega12))
    sph_term = c**2 * alpha12

    # compute integrals for ellipsoidal correction
    k2 = ep2*cos(alpha0)**2

    C40 = (2.0/3 - ep2/15 + 4*ep2**2/105 - 8*ep2**3/315 + 64*ep2**4/3465 - 128*ep2**5/9009) \
        - (1.0/20 - ep2/35 + 2*ep2**2/105 - 16*ep2**3/1155 + 32*ep2**4/3003) * k2 \
        + (1.0/42 - ep2/63 + 8*ep2**2/693 - 90*ep2**3/9009) * k2**2 \
        - (1.0/72 - ep2/99 + 10*ep2**2/1287) * k2**3 \
        + (1.0/110 - ep2/143) * k2**4 - k2**5/156

    C41 = (1.0/180 - ep2/315 + 2*ep2**2/945 - 16*ep2**3/10395 + 32*ep2**4/27027) * k2 \
        - (1.0/252 - ep2/378 + 4*ep2**2/2079 - 40*ep2**3/27027) * k2**2 \
        + (1.0/360 - ep2/495 + 2*ep2**2/1287) * k2**3 \
        - (1.0/495 - 2*ep2/1287) * k2**4 + 5*k2**5/3276

    C42 = (1.0/2100 - ep2/3150 + 4*ep2**2/17325 - 8*ep2**3/45045) * k2**2 \
        - (1.0/1800 - ep2/2475 + 2*ep2**2/6435) * k2**3 \
        + (1.0/1925 - 2*ep2/5005) * k2**4 - k2**5/2184

    C43 = (1.0/17640 - ep2/24255 + 2*ep2**2/63063) * k2**3 \
        - (1.0/10780 - ep2/14014) * k2**4 + 5*k2**5/45864

    C44 = (1.0/124740 - ep2/162162) * k2**4 - 1*k2**5/58968

    C45 = k2**5/792792

    Cs = [C40, C41, C42, C43, C44, C45]
    I4s1 = sum(c*cos((2*i+1)*sigma1) for i,c in enumerate(Cs))
    I4s2 = sum(c*cos((2*i+1)*sigma2) for i,c in enumerate(Cs))

    S12 = sph_term + e**2*a**2 * cos(alpha0)*sin(alpha0) * (I4s2-I4s1)
    return S12


def ellipsoidal_area(a, b, x1, y1, x2, y2):
    """ Area of a single quatrilateral defined by two meridians, the equator,
    and another geodesic.
    """
    _, x1, y1, x2, y2 = _canonical_configuration(x1, y1, x2, y2)
    phi1 = y1*pi/180.0
    phi2 = y2*pi/180.0
    lambda12 = (x2-x1)*pi/180.0

    az, baz, _ = ellipsoidal_inverse(a, b, x1, y1, x2, y2)
    alpha1 = az * pi/180
    alpha2 = (baz-pi) * pi/180
    return _ellipsoidal_area(a, b, lambda12, phi1, phi2, alpha1, alpha2)


###### Utility functions ######
def unroll_angle(alpha):
    if hasattr(alpha, "__len__"):
        alpha = np.array([unroll_angle(a) for a in alpha])
    else:
        while alpha < 0:
            alpha += 2*pi
        while alpha >= 2*pi:
            alpha -= 2*pi
    return alpha

def _degrees(r):
    return r*180/pi

def fzero_brent(a, b, f, tol):
    fa = f(a)
    fb = f(b)
    if fa == 0:
        return a
    elif fb == 0:
        return b
    elif fa * fb > 0:
        raise ValueError("root not bracketed")
    if abs(fa) < abs(fb):
        a,b = b,a
    c = a
    d = np.nan
    mflag = True

    delta = 1e-10
    niter = 0
    while niter != 1000:

        fc = f(c)
        if (fa != fc) and (fb != fc):
            # inverse quadratic interpolation
            s = a*fb*fc / ((fa-fb) * (fa-fc)) \
                    + b*fa*fc / ((fb-fa) * (fb-fc)) \
                    + c*fa*fb / ((fc-fa) * (fc-fb))
        else:
            # secant
            s = b - fb*(b-a)/(fb-fa)

        if ((not 0.25*(3*a+b) < s < b) and (not b < s < 0.25*(3*a+b))) or \
                (mflag and (abs(s-b) >= 0.5*abs(b-c))) or \
                (~mflag and (abs(s-b) >= 0.5*abs(c-d))) or \
                (mflag and (abs(b-c) < abs(delta))) or \
                (~mflag and (abs(c-d) < abs(delta))):
            # bisection
            s = 0.5*(a+b)
            mflag = True
        else:
            mflag = False

        fs = f(s)
        d = c
        c = b
        fa = f(a)
        fb = f(b)
        if fa*fs < 0:
            b = s
            fb = fs
        else:
            a = s
            fa = fs

        if abs(fa) < abs(fb):
            a,b = b,a
            fa,fb = fb,fa

        if fb == 0:
            return b
        elif f(s) == 0:
            return s
        elif (abs(b-a) < tol):
            return b
        niter += 1

    raise ValueError("maximum iterations exceeded (%f, %f)" % (b, s))



from ez_setup import use_setuptools
use_setuptools()
from os.path import exists
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

VERSION = "0.5.0"

class build_ext(_build_ext):

    # solution taken from http://stackoverflow.com/questions/19919905/ \
    #   how-to-bootstrap-numpy-installation-in-setup-py
    def finalize_options(self):
        """ Unsets __NUMPY_SETUP__ so that the get_include function can be used
        """
        _build_ext.finalize_options(self)

        # Hack: unset __NUMPY_SETUP__ so that it can be imported
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

        # Add include_dirs to each extension so that Cython will find them
        for extension in self.extensions:
            extension.include_dirs = self.include_dirs

        # Attempt to compile with Cython - or fallback on local C sources
        try:
            from Cython.Build import cythonize
            for ext in self.extensions:
                sources = []
                for src in ext.sources:
                    if src.endswith(".pyx"):
                        sources.append(src)
                    else:
                        sources.append(src+".pyx")
                ext.sources = sources

            _needs_stub = [ext._needs_stub for ext in self.extensions]
            self.extensions = cythonize(self.extensions)
            for ns, ext in zip(_needs_stub, self.extensions):
                ext._needs_stub = ns

        except ImportError:
            print("Cython not imported")
            print("If C sources exist in the source tree, they will be used")

            for ext in self.extensions:
                ext.sources = list(map(lambda f: f+".c", ext.sources))

        # Check that sources exist for each extension, either from Cython for
        # from having the C-code sitting around. Also, add a _needs_stub
        # attribute to each extension for setuptools.
        for ext in self.extensions:
            if not all(exists(src) for src in ext.sources):
                raise Exception("Extension sources not found - Cython required for install")
        return

# File extension is added to sources at overloaded build_ext.run()
extensions = [Extension("karta.raster.crfuncs", ["karta/raster/crfuncs"]),
              Extension("karta.vector._cvectorgeo", ["karta/vector/_cvectorgeo"])]

setup(
    name = "karta",
    version = VERSION,
    setup_requires = ["numpy>=1.6"],
    install_requires = ["numpy>=1.6", "pyproj>=1.9", "pyshp>=1.2"],
    author = "Nat Wilson",
    author_email = "njwilson23@gmail.com",
    packages = ["karta", "karta.vector", "karta.raster"],
    url = "http://www.ironicmtn.com/karta.html",
    description = "Geospatial analysis in Python",
    long_description = """
*Karta* is a simple and fast framework for spatial analysis in Python.

Create vector geometries:

.. code:: python

    point = Point((-130.0, 52.0), crs=LonLatWGS84)

    line = read_geojson("linedata.json")

    polygon = Polygon([(-515005.78, -1301130.53),
                       (-579174.89, -1282271.94),
                       (-542977.83, -1221147.82),
                       (-437864.05, -1251641.55),
                       (-438160.72, -1252421.48),
                       (-437961.28, -1285314.00)],
                       crs=NSIDCNorth)

Perform simple queries:

.. code:: python

    point2 = Point((-25.0, 48.0), crs=LonLatWGS84)
    point.distance(point2)          # Distance in geographical units

    line.intersects(polygon)        # True or False

    ch = polygon.convex_hull()      # Returns a new polygon
    ch.to_shapefile("poly.shp")

Work with raster data:

.. code:: python

    grid = read_gtiff("landsat_scene.tif")  # Leverages GDAL

    grid.profile(line)              # Collect data along a line

    grid.resample(500.0, 500.0)     # Return a grid resampled at a new resolution

*Karta* works with Python 2.7 and Python 3.3+. Suggestions, bug reports,
test cases, and pull requests are welcome.

DOCUMENTATION
-------------

See the `online
manual <http://www.ironicmtn.com/kartadocs/karta-manual.html>`__, the
`tutorial <http://www.ironicmtn.com/kartadocs/tutorial.html>`__, or read
the `API
documentation <http://www.ironicmtn.com/kartadocs/reference.html>`__.

The manual can also be built offline using Sphinx by running ``make``
from the ``doc/`` subdirectory. The documentation is built from source
code docstrings and the example IPython notebooks, which are also
reproduced in the
`Wiki <https://github.com/njwilson23/karta/wiki/Tutorial>`__. Building
the documentation requires `Sphinx <http://sphinx-doc.org/>`__,
`alabaster <https://github.com/bitprophet/alabaster>`__ and
`numpydoc <https://github.com/numpy/numpydoc>`__.

DEPENDENCIES
------------

Required
~~~~~~~~

-  Python 2.6+ or Python 3.3+
-  numpy
-  pyshp
-  pyproj
-  C-compiler

Recommended
~~~~~~~~~~~

-  osgeom.gdal (for geotiff I/O)
-  osgeo.osr (for coordinate system interchange)
-  scipy

When installing from PyPI, Cython-compiled C source code is provided and
will be automatically compiled to improve performance if a suitable C
compiler is available.
""",
    classifiers = ["Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.2",
                   "Programming Language :: Python :: 3.3",
                   "Programming Language :: Python :: 3.4",
                   "Development Status :: 4 - Beta",
                   "Topic :: Scientific/Engineering",
                   "Topic :: Scientific/Engineering :: GIS",
                   "License :: OSI Approved :: MIT License"],
    license = "MIT License",
    ext_modules = extensions,
    cmdclass = {"build_ext": build_ext},
)




class GeometryError(Exception):
    """ Base class for geometry module errors. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message

class GInitError(GeometryError):
    """ Exception to raise when a geometry object fails to initialize. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message

class GUnitError(GeometryError):
    """ Exception to raise there is a projected unit problem. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message

class GGeoError(GeometryError):
    """ Exception to raise when a geometry object attempts an invalid transform. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message

class CRSError(Exception):
    """ Exception to raise for invalid geodetic operations. """
    def __init__(self, message=''):
        self.message = message

class GridError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class GridIOError(GridError):
    def __init__(self, message=''):
        self.message = message


class NonEquivalentGridError(GridError):
    def __init__(self, A, B, message=''):
        if len(message) == 0:
            self.message = ("{0} and {1} do not share equivalent "
                            "grid layouts".format(A, B))
        else:
            self.message = message

class MissingDependencyError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message




# -*- coding: utf-8 -*-
""" Coordinate reference systems

Implements CRS classes for different kinds of spatial reference systems:

    - CartesianCRS
    - GeographicalCRS
    - SphericalCRS
    - EllipsoidalCRS
    - Proj4CRS

Coordinate reference system (CRS) objects represent mappings geographical
positions to an x, y representation, and contain both projection and geodetic
information. CRS objects should be treated is immutable. It is possible for
multiple CRS objects to implement the same system, and tests for equality may
fail.
"""

import numpy as np
import pyproj
from . import geodesy
from . import errors

try:
    import osgeo.osr
    osgeo.osr.UseExceptions()
    HASOSR = True
except ImportError:
    HASOSR = False

class CRS(object):
    """ Base class for coordinate system instances.

    Subclasses should define:

    - name attribute
    - project(x, y) method
    - forward(x, y, azimuth, distance) method
    - inverse(x0, y0, x1, y1) method

    *project* transforms between world and projected coordinates, and takes the
    optional boolean parameter *inverse* (default `False`)

    *forward* performs a forward geodetic calculation and returns world
    coordinates

    *inverse* performs an inverse geodetic calculation and returns an azimuth,
    a back azimuth, and a distance

    A CRS subclass may optionally also provide string attributes:

    - ref_proj4
    - ref_wkt

    which are used to provide interoperability with other systems.
    """
    def __str__(self):
        return "<CRS {0}>".format(self.name)

    def get_proj4(self):
        if hasattr(self, "ref_proj4"):
            return self.ref_proj4
        else:
            if not HASOSR:
                raise errors.CRSError("no ref_proj4 attribute and no conversion possible (osgeo.osr not installed)")
            srs = osgeo.osr.SpatialReference()
            if hasattr(self, "ref_wkt"):
                srs.ImportFromWkt(self.ref_wkt)
            else:
                raise AttributeError("incomplete CRS definition (missing ref_proj4, ref_wkt attributes)")
            return srs.ExportToProj4()

    def get_wkt(self):
        if hasattr(self, "ref_wkt"):
            return self.ref_proj4
        else:
            if not HASOSR:
                raise errors.CRSError("no ref_wkt attribute and no conversion possible (osgeo.osr not installed)")
            srs = osgeo.osr.SpatialReference()
            if hasattr(self, "ref_proj4"):
                srs.ImportFromProj4(self.ref_proj4.replace("lon", "long"))
            else:
                raise AttributeError("incomplete CRS definition (missing ref_proj4, ref_wkt attributes)")
            return srs.ExportToWkt()

class CartesianCRS(CRS):
    """ Cartesian (flat-earth) reference systems with (x, y) coordinates """
    name = "Cartesian"

    ref_proj4 = ""
    ref_wkt = ""

    @staticmethod
    def project(x, y, inverse=False):
        # Projection on a cartesian coordinate system is the identity
        return x, y

    @staticmethod
    def forward(x, y, az, dist, radians=False):
        """ Forward geodetic problem from a point """
        if not radians:
            az = np.array(az) / 180 * np.pi

        x2 = x + dist * np.sin(az)
        y2 = y + dist * np.cos(az)
        baz = geodesy.unroll_angle(az + np.pi)

        if not radians:
            baz = np.array(baz) * 180 / np.pi
        return x2, y2, baz

    @staticmethod
    def inverse(x1, y1, x2, y2, radians=False):
        """ Inverse geodetic problem to find the geodesic between points """
        dist = geodesy.plane_distance(x1, y1, x2, y2)
        az = geodesy.plane_azimuth(x1, y1, x2, y2)
        baz = geodesy.unroll_angle(az + np.pi)

        if not radians:
            az = az * 180 / np.pi
            baz = baz * 180 / np.pi
        return az, baz, dist

class GeographicalCRS(CRS):
    """ Reference systems with longitude-latitude (θ, φ) coordinates.

    `spheroid` refers to a proj.4 spheroid identifier, e.g. "+ellps=WGS84"
    """
    def __init__(self, spheroid, name):
        self._geod = pyproj.Geod(spheroid)
        self.name = name
        self.ref_proj4 = "+proj=lonlat %s" % self._geod.initstring
        return

    @staticmethod
    def project(x, y, inverse=False, radians=False):
        # Projection on a geographical coordinate system is the identity
        if not radians:
            return x, y
        else:
            return np.array(x)/180 * np.pi, np.array(y)/180 * np.pi

    def forward(self, *args, **kwargs):
        """ Forward geodetic problem from a point """
        return self._geod.fwd(*args, **kwargs)

    def inverse(self, *args, **kwargs):
        """ Inverse geodetic problem to find the geodesic between points """
        return self._geod.inv(*args, **kwargs)

class SphericalCRS(GeographicalCRS):
    """ Spherical geographic coordinate system defined by a radius. """
    name = "Spherical"

    def __init__(self, radius):
        self.radius = radius
        return

    @property
    def ref_proj4(self):
        return "+proj=lonlat +a=%f +b=%f +no_defs" % (self.radius, self.radius)

    @property
    def ref_wkt(self):
        return ('GEOGCS["unnamed ellipse",DATUM["unknown",'
                'SPHEROID["unnamed",%f,0]],PRIMEM["Greenwich",0],'
                'UNIT["degree",0.0174532925199433]]' % self.radius)

    def forward(self, lons, lats, az, dist, radians=False):
        """ Forward geodetic problem from a point """
        if not radians:
            lons = np.array(lons) * np.pi / 180.0
            lats = np.array(lats) * np.pi / 180.0
            az = np.array(az) * np.pi / 180.0

        d_ = dist / self.radius
        lats2 = np.arcsin(np.sin(lats) * np.cos(d_) +
                    np.cos(lats) * np.sin(d_) * np.cos(az))
        dlons = np.arccos((np.cos(d_) - np.sin(lats2) * np.sin(lats)) /
                          (np.cos(lats) * np.cos(lats2)))
        baz = np.arccos((np.sin(lats) - np.cos(d_) * np.sin(lats2)) /
                        (np.sin(d_) * np.cos(lats2)))
        if 0 <= az < np.pi:
            lons2 = lons + dlons
            baz = -baz
        elif np.pi <= az < 2*np.pi:
            lons2 = lons - dlons
        else:
            raise ValueError("azimuth should be [0, 2pi)")

        baz = geodesy.unroll_angle(baz)

        if not radians:
            lons2 = np.array(lons2) * 180 / np.pi
            lats2 = np.array(lats2) * 180 / np.pi
            baz = np.array(baz) * 180 / np.pi
        return lons2, lats2, baz

    def inverse(self, lons1, lats1, lons2, lats2, radians=False):
        """ Inverse geodetic problem to find the geodesic between points """
        if not radians:
            lons1 = lons1 * np.pi / 180.0
            lons2 = lons2 * np.pi / 180.0
            lats1 = lats1 * np.pi / 180.0
            lats2 = lats2 * np.pi / 180.0

        az = geodesy.sphere_azimuth(lons1, lats1, lons2, lats2)
        baz = geodesy.sphere_azimuth(lons2, lats2, lons1, lats1)
        dist = geodesy.sphere_distance(lons1, lats1, lons2, lats2, self.radius)

        if not radians:
            az = az * 180 / np.pi
            baz = baz * 180 / np.pi
        return az, baz, dist

class EllipsoidalCRS(GeographicalCRS):
    """ Ellipsoidal geographic coordinate system defined by equatorial and
    polar radii.
    """
    name = "Ellipsoidal"

    def __init__(self, a, b):
        """ Define a geographical coordinate system on an ellipse. *a* is the
        equatorial radius, and *b* is the polar radius. """
        self.a = a
        self.b = b
        self.ref_proj4 = "+proj=lonlat +units=m +a=%f +b=%f +no_defs" % (a, b)
        return

    def get_proj4(self):
        return self.ref_proj4

    def project(self, x, y, **kw):
        # Projection on a geographical coordinate system is the identity
        return x, y

    def forward(self, x, y, azimuth, distance, radians=False):
        """ Forward geodetic problem from a point """
        if radians:
            x *= 180.0/pi
            y *= 180.0/pi
            azimuth *= 180.0/pi

        x2, y2, baz = geodesy.ellipsoidal_forward(self.a, self.b, x, y, azimuth, distance)

        if radians:
            x2 *= pi/180.0
            y2 *= pi/180.0
            baz *= pi/180.0

        return x2, y2, baz

    def inverse(self, x1, y1, x2, y2, radians=False):
        """ Inverse geodetic problem to find the geodesic between points """
        if radians:
            x1 *= 180.0/pi
            y1 *= 180.0/pi
            x2 *= 180.0/pi
            y2 *= 180.0/pi

        az, baz, dist = geodesy.ellipsoidal_inverse(self.a, self.b, x1, y1, x2, y2)

        if radians:
            az *= pi/180.0
            baz *= pi/180.0

        return az, baz, dist

class Proj4CRS(CRS):
    """ Custom reference systems, which may be backed by a *pypoj.Proj* instance
    or a custom projection function.

    `proj` is a dictionary used to define an instance of `pyproj.Proj`

    `spheroid` refers to a proj.4 spheroid identifier, e.g. "+ellps=WGS84"
    """
    def __init__(self, proj, spheroid=None, name=None):
        if spheroid is None:
            # Extract the spheroid from the +ellps, +datum, or +a,+b parameters
            # if present
            if "+ellps" in proj:
                for kv in proj.split():
                    if "+ellps" in kv:
                        k,v = kv.split("=")
                        spheroid = "+ellps=%s" % v
                        break
            elif "+datum" in proj:
                for kv in proj.split():
                    if "+datum" in kv:
                        k,v = kv.split("=")
                        spheroid = "+ellps=%s" % v
                        break
            elif ("+a" in proj) and ("+b" in proj):
                a, b = None, None
                for kv in proj.split():
                    if "+a" in kv:
                        _,a = kv.split("=")
                    elif "+b" in kv:
                        _,b = kv.split("=")
                    if None not in (a,b):
                        spheroid = "+a=%s +b=%s" % (a, b)
                        break
            else:
                raise errors.CRSError("Spheroid must be provided: %s" % proj)

        self.project = pyproj.Proj(proj)
        self._geod = pyproj.Geod(spheroid)

        self.initstring_proj = self.project.srs
        self.initstring_geod = self._geod.initstring
        self.ref_proj4 = "%s %s" % (self.project.srs, self._geod.initstring)

        if name is not None:
            self.name = name
        else:
            self.name = proj
        return

    def __eq__(self, other):
        if (getattr(self.project, "srs", 0) == getattr(other.project, "srs", 1) and
            getattr(self._geod, "initstring", 0) == getattr(other._geod, "initstring", 1)):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def forward(self, *args, **kwargs):
        """ Forward geodetic problem from a point """
        return self._geod.fwd(*args, **kwargs)

    def inverse(self, *args, **kwargs):
        """ Inverse geodetic problem to find the geodesic between points """
        return self._geod.inv(*args, **kwargs)

def crs_from_wkt(wkt):
    srs = osgeo.osr.SpatialReference(wkt)
    srs.MorphFromESRI()     # munge parameters for proj.4 export
    if srs.IsGeographic():
        proj4 = srs.ExportToProj4()
        for arg in proj4.split():
            if "+ellps" in arg:
                return GeographicalCRS(arg, "unnamed")
            if "+datum" in arg:
                _,v = arg.split("=")
                return GeographicalCRS("+ellps=%s" % v, "unnamed")
        raise errors.CRSError("Could not interpret %s as geographical CRS" % proj4)
    else:
        return Proj4CRS(srs.ExportToProj4())

############ Predefined CRS instances ############

Cartesian = CartesianCRS()
SphericalEarth = SphericalCRS(6371009.0)
LonLatWGS84 = EllipsoidalCRS(6378137.0, 6356752.314245)
LonLatNAD83 = EllipsoidalCRS(6378137.0, 6356752.314140)
LonLatNAD27 = EllipsoidalCRS(6378206.4, 6356583.8)

LonLatWGS84_proj4 = GeographicalCRS("+ellps=WGS84", "WGS84 (Geographical)")
LonLatNAD27_proj4 = GeographicalCRS("+ellps=clrk66", "NAD27 (Geographical)")
LonLatNAD83_proj4 = GeographicalCRS("+ellps=GRS80", "NAD83 (Geographical)")

UPSNorth = Proj4CRS(proj="+proj=stere +lat_0=90 +lat_ts=90 +lon_0=0 +k=0.994 +x_0=2000000 +y_0=2000000 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="Universal Polar Stereographic (North)")

UPSSouth = Proj4CRS(proj="+proj=stere +lat_0=-90 +lat_ts=-90 +lon_0=0 +k=0.994 +x_0=2000000 +y_0=2000000 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="Universal Polar Stereographic (South)")

NSIDCNorth = Proj4CRS(proj="+proj=stere +lat_0=90 +lat_ts=70 +lon_0=-45 +k=1 +x_0=0 +y_0=0 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="NSIDC (North)")

NSIDCSouth = Proj4CRS(proj="+proj=stere +lat_0=-90 +lat_ts=-70 +lon_0=0 +k=1 +x_0=0 +y_0=0 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="NSIDC (South)")

LambertEqualArea = Proj4CRS(proj="+proj=laea +lat_0=0 +lon_0=0 +x_0=0 +y_0=0",
        spheroid="+ellps=WGS84", name="Lambert Equal Area")

GallPetersEqualArea = Proj4CRS("+proj=cea +lon_0=0 +lat_ts=45 +x_0=0 +y_0=0 "
                               "+ellps=WGS84 +units=m +no_defs",
        spheroid="+ellps=WGS84", name="Gall Peters Equal Area")

WebMercator = Proj4CRS("+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 "
                       "+lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m "
                       "+nadgrids=@null +wktext +no_defs",
        spheroid="+a=6378137 +b=6378137", name="Web Mercator")

